#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import os
from abc import ABC, abstractmethod
from copy import copy
from functools import wraps
from typing import TYPE_CHECKING, Union, Dict
from warnings import warn
from contextlib import redirect_stdout

from pandas import DataFrame
import ibm_watson_machine_learning._wrappers.requests as requests

from ..experiment import AutoAI
from ..utils import is_lale_pipeline
from ..utils.autoai.utils import (prepare_auto_ai_model_to_publish, prepare_auto_ai_model_to_publish_normal_scenario,
                                  remove_file, prepare_auto_ai_model_to_publish_notebook,
                                  prepare_auto_ai_model_to_publish_notebook_normal_scenario,
                                  download_wml_pipeline_details_from_file, get_sw_spec_and_type_based_on_sklearn,
                                  check_if_ts_pipeline_is_winner, convert_dataframe_to_fields_values_payload)
from ..utils.deployment.errors import (WrongDeploymnetType, ModelTypeNotSupported, NotAutoAIExperiment,
                                       DeploymentNotSupported, MissingSpace, ModelPromotionFailed,
                                       ServingNameNotAvailable)
from ..helpers import DataConnection
from ..wml_client_error import ApiRequestFailure
from ..workspace import WorkSpace

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline
    from numpy import ndarray

__all__ = [
    "BaseDeployment"
]


class BaseDeployment(ABC):
    """Base abstract class for Deployment."""

    def __init__(self,
                 deployment_type: str,
                 source_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 source_project_id: str = None,
                 source_space_id: str = None,
                 target_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 target_project_id: str = None,
                 target_space_id: str = None,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None):

        if space_id is None and source_space_id is None and target_space_id is None:
            raise MissingSpace(reason="Any of the [space_id, source_space_id, target_space_id] is not specified.")

        # note: backward compatibility
        if wml_credentials is not None:
            source_wml_credentials = wml_credentials
            warn("\"wml_credentials\" parameter is deprecated, please use \"source_wml_credentials\"")
            print("\"wml_credentials\" parameter is deprecated, please use \"source_wml_credentials\"")

        if project_id is not None:
            source_project_id = project_id
            warn("\"project_id\" parameter is deprecated, please use \"source_project_id\"")
            print("\"project_id\" parameter is deprecated, please use \"source_project_id\"")

        if space_id is not None:
            source_space_id = space_id
            warn("\"space_id\" parameter is deprecated, please use \"source_space_id\"")
            print("\"space_id\" parameter is deprecated, please use \"source_space_id\"")
        # --- end note

        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        if isinstance(source_wml_credentials, WorkSpace):
            self._source_workspace = source_wml_credentials

        elif isinstance(source_wml_credentials, dict):
            self._source_workspace = WorkSpace(wml_credentials=source_wml_credentials.copy(),
                                               project_id=source_project_id,
                                               space_id=source_space_id)
        else:
            self._source_workspace = None

        if target_wml_credentials is None:
            if isinstance(source_wml_credentials, WorkSpace):
                self._target_workspace = WorkSpace(wml_credentials=source_wml_credentials.wml_credentials,
                                                   space_id=target_space_id if target_space_id is not None else source_wml_credentials.space_id)

            elif isinstance(source_wml_credentials, dict):
                self._target_workspace = WorkSpace(wml_credentials=source_wml_credentials.copy(),
                                                   space_id=target_space_id if target_space_id is not None else source_space_id)
            else:
                self._target_workspace = None

            if target_space_id or target_project_id:
                self._target_workspace.project_id = target_project_id
                self._target_workspace.space_id = target_space_id

        else:
            if isinstance(target_wml_credentials, WorkSpace):
                self._target_workspace = target_wml_credentials

            elif isinstance(target_wml_credentials, dict):
                self._target_workspace = WorkSpace(wml_credentials=target_wml_credentials.copy(),
                                                   project_id=target_project_id,
                                                   space_id=target_space_id)

                # note: only if user provides target WML information
                if self._source_workspace is None:
                    self._source_workspace = copy(self._target_workspace)

            else:
                self._target_workspace = None
        # --- end note

        if self._source_workspace is None and self._target_workspace is None:
            raise TypeError(f"{self.__class__.__name__} is missing one of the parameters: "
                            f"['source_wml_credentials', 'target_wml_credentials']")

        self.name = None
        self.id = None
        if deployment_type == 'online':
            self.scoring_url = None

        if deployment_type == 'batch':
            self._obm = False

    def __repr__(self):
        return f"name: {self.name}, id: {self.id}"

    def __str__(self):
        return f"name: {self.name}, id: {self.id}"

    @abstractmethod
    def create(self, **kwargs):
        """Create deployment from a model.

        :param model: AutoAI model name
        :type model: str

        :param deployment_name: name of the deployment
        :type deployment_name: str

        :param training_data: training data for the model
        :type training_data: pandas.DataFrame or numpy.ndarray, optional

        :param training_target: target/label data for the model
        :type training_target: pandas.DataFrame or numpy.ndarray, optional

        :param metadata: model meta properties
        :type metadata: dict, optional

        :param experiment_run_id: ID of a training/experiment (only applicable for AutoAI deployments)
        :type experiment_run_id: str, optional
        """

        if kwargs.get('serving_name'):
            status_code, response = self._target_workspace.wml_client.deployments._get_serving_name_info(kwargs.get('serving_name'))

            if status_code == 409:
                raise ServingNameNotAvailable(response['errors'][0])

        # note: This section is only for deployments with specified experiment_id
        if kwargs['experiment_run_id'] is not None:
            run_params = self._source_workspace.wml_client.training.get_details(
                training_uid=kwargs['experiment_run_id'], _internal=True)
            wml_pipeline_details = self._source_workspace.wml_client.pipelines.get_details(
                run_params['entity']['pipeline']['id'])

            if not ('autoai' in str(wml_pipeline_details) or 'auto_ai' in str(wml_pipeline_details)):
                raise NotAutoAIExperiment(
                    kwargs['experiment_run_id'], reason="Currently WebService class supports only AutoAI models.")

            if 'auto_ai.obm' in str(wml_pipeline_details):
                if hasattr(self, 'scoring_url'):
                    raise DeploymentNotSupported('WebService',
                                                 reason="AutoAI with DataJoin is not supported for WebService "
                                                        "deployment. Please use Batch deployment instead.")
                self._obm = True

            else:
                self._obm = False

            print("Preparing an AutoAI Deployment...")
            # TODO: remove part with model object depployment
            if not isinstance(kwargs['model'], str):
                warning_msg = ("Depreciation Warning: Passing an object will no longer be supported. "
                               "Please specify the AutoAI model name to deploy.")
                print(warning_msg)
                warn(warning_msg)

            if is_lale_pipeline(kwargs['model']):
                model = kwargs['model'].export_to_sklearn_pipeline()
            else:
                model = kwargs['model']

            # note: check if model is of lale type, if yes, convert it back to scikit
            if not isinstance(kwargs['model'], str):
                model_props = {
                    self._target_workspace.wml_client.repository.ModelMetaNames.NAME: f"{kwargs['deployment_name']} Model",
                    self._target_workspace.wml_client.repository.ModelMetaNames.TYPE: "wml-hybrid_0.1",
                    self._target_workspace.wml_client.repository.ModelMetaNames.SOFTWARE_SPEC_UID:
                        self._target_workspace.wml_client.software_specifications.get_uid_by_name("hybrid_0.1")
                }

                # fill additional needed OBM metadata
                if self._obm:
                    model_props[self._target_workspace.wml_client.repository.ModelMetaNames.TRAINING_DATA_REFERENCES] = (
                        run_params['entity']['training_data_references']
                    )
                # --- end note

                schema, artifact_name = prepare_auto_ai_model_to_publish(
                    pipeline_model=model,
                    run_params=run_params,
                    run_id=kwargs['experiment_run_id'],
                    wml_client=self._source_workspace.wml_client)

                model_props[self._target_workspace.wml_client.repository.ModelMetaNames.INPUT_DATA_SCHEMA] = (
                    schema['schemas'] if self._obm else [schema])

            else:
                # raise an error when TS pipeline is discarded one
                check_if_ts_pipeline_is_winner(details=run_params, model_name=model)

                # Note: We need to fetch credentials when 'container' is the type
                if run_params['entity']['results_reference']['type'] == 'container':
                    data_connection = DataConnection._from_dict(_dict=run_params['entity']['results_reference'])
                    data_connection._wml_client = self._source_workspace.wml_client
                else:
                    data_connection = None
                # --- end note

                try:
                    auto_pipelines_parameters = wml_pipeline_details.get('entity', {}).get('document', {}).get('pipelines', [])[0].get('nodes', [])[0].get('parameters')
                except:
                    auto_pipelines_parameters = None

                artifact_name, model_props = prepare_auto_ai_model_to_publish_normal_scenario(
                    pipeline_model=model,
                    run_params=run_params,
                    run_id=kwargs['experiment_run_id'],
                    wml_client=self._source_workspace.wml_client,
                    space_id=self._target_workspace.space_id,
                    result_reference=data_connection,
                    auto_pipelines_parameters=auto_pipelines_parameters
                )

            deployment_details = self._deploy(
                pipeline_model=artifact_name,
                deployment_name=kwargs['deployment_name'],
                serving_name=kwargs.get('serving_name'),
                meta_props=model_props,
                hardware_spec=kwargs.get('hardware_spec')
            )

            remove_file(filename=artifact_name)

            self.name = kwargs['deployment_name']
            self.id = deployment_details.get('metadata').get('id')
            if kwargs['deployment_type'] == 'online':
                self.scoring_url = self._target_workspace.wml_client.deployments.get_scoring_href(deployment_details)
        # --- end note

        # note: This section is for deployments from auto-gen notebook with COS connection / WSD
        else:
            # note: only if we have COS connections from the notebook or for WSD
            if kwargs.get('metadata') is not None:
                print("Preparing an AutoAI Deployment...")

                # note: WMLS
                if self._source_workspace is not None and self._source_workspace.WMLS:
                    optimizer = AutoAI(
                        self._source_workspace.wml_credentials,
                        space_id=self._source_workspace.space_id
                    ).runs.get_optimizer(metadata=kwargs['metadata'])

                # note: CP4D 3.5
                elif self._source_workspace is not None and self._source_workspace.wml_client.ICP:
                    optimizer = AutoAI(self._source_workspace).runs.get_optimizer(metadata=kwargs['metadata'])

                # note: CLOUD
                else:
                    optimizer = AutoAI().runs.get_optimizer(metadata=kwargs['metadata'],
                                                            wml_client=self._source_workspace.wml_client)

                # note: check for obm step in pipeline details
                if hasattr(optimizer, '_result_client'):
                    wml_pipeline_details = download_wml_pipeline_details_from_file(optimizer._result_client)

                else:
                    run_id = optimizer._engine._current_run_id
                    pipeline_id = optimizer._workspace.wml_client.training.get_details(
                        run_id, _internal=True)['entity']['pipeline']['id']
                    wml_pipeline_details = optimizer._workspace.wml_client.pipelines.get_details(pipeline_id)

                if 'auto_ai.obm' in str(wml_pipeline_details):
                    if hasattr(self, 'scoring_url'):
                        raise DeploymentNotSupported('WebService',
                                                     reason="AutoAI with DataJoin is not supported for WebService "
                                                            "deployment. Please use Batch deployment instead.")
                    self._obm = True

                else:
                    self._obm = False
                # --- end note

                # note: only when user did not pass WMLS credentials during Service initialization
                if self._source_workspace is None:
                    self._source_workspace = copy(optimizer._workspace)

                if self._target_workspace is None:
                    self._target_workspace = copy(optimizer._workspace)
                # --- end note

                # TODO: remove part with model object depployment
                if not isinstance(kwargs['model'], str):
                    warning_msg = ("Depreciation Warning: Passing an object will no longer be supported. "
                                   "Please specify the AutoAI model name to deploy.")
                    print(warning_msg)
                    warn(warning_msg)

                if is_lale_pipeline(kwargs['model']):
                    model = kwargs['model'].export_to_sklearn_pipeline()
                else:
                    model = kwargs['model']

                if not isinstance(kwargs['model'], str):
                    model_props = {
                        self._target_workspace.wml_client.repository.ModelMetaNames.NAME: f"{kwargs['deployment_name']} Model",
                        self._target_workspace.wml_client.repository.ModelMetaNames.TYPE: "wml-hybrid_0.1",
                        self._target_workspace.wml_client.repository.ModelMetaNames.SOFTWARE_SPEC_UID:
                            self._target_workspace.wml_client.software_specifications.get_uid_by_name("hybrid_0.1")
                    }

                    # note: CLOUD
                    if not self._target_workspace.wml_client.ICP:
                        schema, artifact_name = prepare_auto_ai_model_to_publish_notebook(
                            pipeline_model=model,
                            result_connection=optimizer._result_client[0],
                            cos_client=optimizer._result_client[1],
                            obm=True if self._obm else False
                        )

                        model_props[self._target_workspace.wml_client.repository.ModelMetaNames.INPUT_DATA_SCHEMA] = (
                            schema['schemas'] if self._obm else [schema])

                        deployment_details = self._deploy(
                            pipeline_model=artifact_name,
                            deployment_name=kwargs['deployment_name'],
                            meta_props=model_props,
                            serving_name=kwargs.get('serving_name'),
                            result_client=optimizer._result_client
                        )

                    # note: WSD part
                    else:
                        training_result_reference = kwargs['metadata'].get('training_result_reference')
                        run_id = training_result_reference.location.path.split('/')[-3]
                        run_params = self._source_workspace.wml_client.training.get_details(training_uid=run_id, _internal=True)

                        schema, artifact_name = prepare_auto_ai_model_to_publish(
                            pipeline_model=model,
                            run_params=run_params,
                            run_id=kwargs['experiment_run_id'],
                            wml_client=self._source_workspace.wml_client)

                        model_props[self._target_workspace.wml_client.repository.ModelMetaNames.INPUT_DATA_SCHEMA] = (
                            schema['schemas'] if self._obm else [schema])

                        deployment_details = self._deploy(
                            pipeline_model=artifact_name,
                            deployment_name=kwargs['deployment_name'],
                            serving_name=kwargs.get('serving_name'),
                            meta_props=model_props
                        )
                    # --- end note

                else:
                    if not self._target_workspace.wml_client.ICP:

                        optimizer_2 = AutoAI(self._source_workspace).runs.get_optimizer(metadata=kwargs['metadata'])
                        run_details = optimizer_2._workspace.wml_client.training.get_details(
                            optimizer_2.get_params()['run_id'], _internal=True
                        )

                        artifact_name, model_props = prepare_auto_ai_model_to_publish_notebook_normal_scenario(
                            pipeline_model=model,
                            result_connection=optimizer._result_client[0],
                            cos_client=optimizer._result_client[1],
                            run_params=run_details,
                            space_id=self._target_workspace.space_id,
                            auto_pipelines_parameters=optimizer.get_params()
                        )

                        deployment_details = self._deploy(
                            pipeline_model=artifact_name,
                            deployment_name=kwargs['deployment_name'],
                            meta_props=model_props,
                            serving_name=kwargs.get('serving_name'),
                            result_client=optimizer._result_client
                        )

                    # note: WSD / CP4D 3.5 part
                    else:
                        training_result_reference = kwargs['metadata'].get('training_result_reference')
                        run_id = training_result_reference.location.path.split('/')[-3]
                        run_params = self._source_workspace.wml_client.training.get_details(training_uid=run_id, _internal=True)

                        artifact_name, model_props = prepare_auto_ai_model_to_publish_normal_scenario(
                            pipeline_model=model,
                            run_params=run_params,
                            run_id=run_id,
                            wml_client=self._source_workspace.wml_client,
                            space_id=self._target_workspace.space_id,
                            auto_pipelines_parameters=optimizer.get_params()
                        )

                        deployment_details = self._deploy(
                            pipeline_model=artifact_name,
                            deployment_name=kwargs['deployment_name'],
                            serving_name=kwargs.get('serving_name'),
                            meta_props=model_props
                        )
                    # --- end note

                remove_file(filename=artifact_name)

                self.name = kwargs['deployment_name']
                self.id = deployment_details.get('metadata').get('id')
                if kwargs['deployment_type'] == 'online':
                    self.scoring_url = self._target_workspace.wml_client.deployments.get_scoring_href(
                        deployment_details
                    )
            # --- end note

            else:
                raise ModelTypeNotSupported(type(kwargs['model']),
                                            reason="Currently WebService class supports only AutoAI models.")

    @abstractmethod
    def get_params(self):
        """Get deployment parameters."""
        return self._target_workspace.wml_client.deployments.get_details(self.id)

    @abstractmethod
    def score(self, **kwargs):
        """Scoring on WML. Payload is passed to the WML scoring endpoint where model have been deployed.

        :param payload: data to test the model
        :type payload: pandas.DataFrame
        """

        if isinstance(kwargs['payload'], DataFrame):
            payload = convert_dataframe_to_fields_values_payload(kwargs['payload'])
            input_data = [payload]

        elif isinstance(kwargs['payload'], dict):
            observations_df = kwargs['payload'].get('observations', DataFrame())
            supporting_features_df = kwargs['payload'].get('supporting_features')

            if isinstance(observations_df, DataFrame) and \
                    (isinstance(supporting_features_df, DataFrame) or supporting_features_df is None):
                observations_payload = convert_dataframe_to_fields_values_payload(observations_df)
                observations_payload['id'] = 'observations'
                input_data = [observations_payload]

                if supporting_features_df is not None:
                    supporting_features_payload = convert_dataframe_to_fields_values_payload(supporting_features_df)
                    supporting_features_payload['id'] = 'supporting_features'
                    input_data.append(supporting_features_payload)
            else:
                raise TypeError('Missing data observations in payload or observations'
                                 'or supporting_features are not pandas DataFrames.')

        else:
            raise TypeError(f"Scoring payload has invalid type: {type(kwargs['payload'])}. "
                            f"Supported types are: pandas.DataFrame and dictionary.")

        transaction_id = kwargs.get('transaction_id')

        scoring_payload = {self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA: input_data}

        score = self._target_workspace.wml_client.deployments.score(self.id, scoring_payload, transaction_id=transaction_id)

        return score

    @abstractmethod
    def delete(self, **kwargs):
        """Delete deployment on WML.

        :param deployment_id: ID of the deployment to delete, if empty, current deployment will be deleted
        :type deployment_id: str, optional

        :param deployment_type: type of the deployment: [online, batch]
        :type deployment_type: str
        """
        if kwargs['deployment_id'] is None:
            self._target_workspace.wml_client.deployments.delete(self.id)
            self.name = None
            self.scoring_url = None
            self.id = None

        else:
            deployment_details = self._target_workspace.wml_client.deployments.get_details(
                deployment_uid=kwargs['deployment_id'])
            if deployment_details.get('entity', {}).get(kwargs['deployment_type']) is not None:
                self._target_workspace.wml_client.deployments.delete(kwargs['deployment_id'])

            else:
                raise WrongDeploymnetType(
                    f"{kwargs['deployment_type']}",
                    reason=f"Deployment with ID: {kwargs['deployment_id']} is not of \"{kwargs['deployment_type']}\" type!")

    @abstractmethod
    def list(self, **kwargs):
        """List WML deployments.

        :param limit: set the limit of how many deployments to list,
            default is `None` (all deployments should be fetched)
        :type limit: int, optional

        :param deployment_type: type of the deployment: [online, batch]
        :type deployment_type: str
        """
        deployments = self._target_workspace.wml_client.deployments.get_details(limit=kwargs['limit'])
        columns = [
            'created_at',
            'modified_at',
            'id',
            'name',
            'status'
        ]

        data = [
            [deployment.get('metadata')['created_at'],
             deployment.get('metadata')['modified_at'],
             deployment.get('metadata')['id'],
             deployment.get('metadata')['name'],
             deployment.get('entity')['status']['state'],
             ] for deployment in deployments.get('resources', []) if
            isinstance(deployment.get('entity', {}).get(kwargs['deployment_type']), dict)
        ]

        deployments = DataFrame(data=data, columns=columns).sort_values(by=['created_at'], ascending=False)
        return deployments.head(n=kwargs['limit'])

    @abstractmethod
    def get(self, **kwargs):
        """Get WML deployment.

        :param deployment_id: ID of the deployment to work with
        :type deployment_id: str

        :param deployment_type: type of the deployment: [online, batch]
        :type deployment_type: str
        """
        deployment_details = self._target_workspace.wml_client.deployments.get_details(
            deployment_uid=kwargs['deployment_id'])
        if deployment_details.get('entity', {}).get(kwargs['deployment_type']) is not None:
            self.name = deployment_details.get('metadata').get('name')
            self.id = deployment_details.get('metadata').get('id')
            if kwargs['deployment_type'] == 'online':
                self.scoring_url = self._target_workspace.wml_client.deployments.get_scoring_href(deployment_details)

        else:
            raise WrongDeploymnetType(
                f"{kwargs['deployment_type']}",
                reason=f"Deployment with ID: {kwargs['deployment_id']} is not of \"{kwargs['deployment_type']}\" type!")

    @abstractmethod
    def _deploy(self, **kwargs):
        """Protected method to create a deployment."""
        pass

    def _publish_model(self,
                       pipeline_model: Union['Pipeline', str],
                       meta_props: Dict) -> str:
        """Publish model into WML.

        :param pipeline_model: model of the pipeline to publish
        :type pipeline_model: Pipeline or str

        :param meta_props: model meta properties
        :type meta_props: dict

        :return: asset id
        :rtype: str
        """

        # Note: publish model to project and then promote to space
        if self._source_workspace.project_id and self._source_workspace.project_id in str(meta_props):
            published_model_details = self._source_workspace.wml_client.repository.store_model(
                model=pipeline_model,
                meta_props=meta_props)

            project_asset_id = self._source_workspace.wml_client.repository.get_model_id(published_model_details)

            asset_id = self._source_workspace.wml_client.repository.promote_model(model_id=project_asset_id,
                                                                                  source_project_id=self._source_workspace.project_id,
                                                                                  target_space_id=self._target_workspace.space_id)

        else:
            published_model_details = self._target_workspace.wml_client.repository.store_model(
                model=pipeline_model,
                meta_props=meta_props)

            asset_id = self._target_workspace.wml_client.repository.get_model_id(published_model_details)

        print(f"Published model uid: {asset_id}")
        return asset_id

    @staticmethod
    def _project_to_space_to_project(method):
        @wraps(method)
        def _method(self, *method_args, **method_kwargs):
            with redirect_stdout(open(os.devnull, "w")):
                if self._target_workspace.space_id is not None:
                    self._target_workspace.wml_client.set.default_space(
                        self._target_workspace.space_id) if self._target_workspace.wml_client.ICP else None

            try:
                return method(self, *method_args, **method_kwargs)

            finally:
                if not self._target_workspace.WMLS and self._target_workspace.project_id:
                    with redirect_stdout(open(os.devnull, "w")):
                        if self._target_workspace.project_id is not None:
                            self._target_workspace.wml_client.set.default_project(
                                self._target_workspace.project_id) if self._target_workspace.wml_client.ICP else None

        return _method



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------


import sys
import pandas as pd


class LocalBatchReader:
    """LocalBatchReader is designed to """
    def __init__(self, file_path: str, batch_size: int = 1073741824 // 10):
        self.file_path = file_path
        self.batch_size = batch_size    # default 100 MB
        self.row_size = 0

        self._determine_row_size()

    def _determine_row_size(self) -> None:
        data_row = pd.read_csv(self.file_path, chunksize=1)
        self.row_size = sys.getsizeof(data_row)

    def _calculate_chunk_size(self) -> int:
        return self.batch_size // self.row_size

    def __iter__(self) -> pd.DataFrame:
        x = self._calculate_chunk_size()
        return pd.read_csv(self.file_path, chunksize=self._calculate_chunk_size())



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from typing import TYPE_CHECKING, List, Union

from ibm_watson_machine_learning.helpers.connections import (
    DataConnection, S3Location, FSLocation, AssetLocation, CP4DAssetLocation, WMLSAssetLocation, CloudAssetLocation,
    WSDAssetLocation, DeploymentOutputAssetLocation, ContainerLocation, S3Connection)
from ibm_watson_machine_learning.utils.autoai.enums import (DataConnectionTypes)
from ibm_watson_machine_learning.utils.autoai.errors import ContainerTypeNotSupported

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace


__all__ = [
    "validate_source_data_connections",
    "create_results_data_connection",
    "validate_results_data_connection",
    "create_deployment_output_data_connection",
    "validate_deployment_output_connection"
]


def validate_source_data_connections(source_data_connections: List['DataConnection'],
                                     workspace: 'WorkSpace',
                                     deployment=False) -> List['DataConnection']:
    for data_connection in source_data_connections:

        if isinstance(data_connection.location, FSLocation):
            # note: save data as an data asset
            if workspace.wml_client.ICP:
                asset_id = data_connection.location._save_file_as_data_asset(workspace=workspace)
                data_connection.location = AssetLocation(asset_id)
                data_connection.type = DataConnectionTypes.DS
            # --- end note

        elif isinstance(data_connection.location, WSDAssetLocation):
            # note: try to move local asset to server
            data_connection.location._move_asset_to_server(workspace=workspace)
            # --- end note
        elif isinstance(data_connection.location, AssetLocation) and data_connection.location._wsd:
            # note: try to move local asset to server
            data_connection.location._wsd_move_asset_to_server(workspace=workspace)
            # --- end note
        elif isinstance(data_connection.location, ContainerLocation):
            if workspace.wml_client.ICP:
                raise ContainerTypeNotSupported() # block Container type on CPD
            elif isinstance(data_connection.connection, S3Connection):
                # note: remove S3 inline credential from data asset before training
                data_connection.connection = None
                if hasattr(data_connection.location, 'bucket'):
                    delattr(data_connection.location, 'bucket')
                # --- end note

        elif isinstance(data_connection.connection, S3Connection) and isinstance(data_connection.location,
                                                                                 AssetLocation):
            # note: remove S3 inline credential from data asset before training
            data_connection.connection = None

            for s3_attr in ['bucket', 'path']:
                if hasattr(data_connection.location, s3_attr):
                    delattr(data_connection.location, s3_attr)
            # --- end note

    return source_data_connections


def create_results_data_connection(source_data_connections: List['DataConnection'],
                                   workspace: 'WorkSpace') -> 'DataConnection':
    if isinstance(source_data_connections[0].location, S3Location):
        results_data_connection = DataConnection(
            connection=source_data_connections[0].connection,
            location=S3Location(bucket=source_data_connections[0].location.bucket,
                                path='.')
        )
    else:
        location = FSLocation()
        if workspace.wml_client.default_space_id:
            location.path = location.path.format(option='spaces',
                                                 id=workspace.wml_client.default_space_id)
        else:
            location.path = location.path.format(option='projects',
                                                 id=workspace.wml_client.default_project_id)
        results_data_connection = DataConnection(
            connection=None,
            location=location
        )
    return results_data_connection


def validate_results_data_connection(results_data_connection: Union['DataConnection', None],
                                     workspace: 'WorkSpace',
                                     source_data_connections: List['DataConnection'] = None) -> 'DataConnection':
    # note: if user did not provide results storage information, use default ones
    if results_data_connection is None and source_data_connections is not None:
        results_data_connection = create_results_data_connection(source_data_connections=source_data_connections,
                                                                 workspace=workspace)
    # -- end note
    # note: results can be stored only on FS or COS
    if not isinstance(results_data_connection.location, (S3Location, FSLocation)):
        raise TypeError('Unsupported results location type. Results reference can be stored'
                        ' only on S3Location or FSLocation.')
    # -- end
    return results_data_connection


def create_deployment_output_data_connection(source_data_connections: List['DataConnection'],
                                             output_filename = 'deployment_output.csv') -> 'DataConnection':
    if isinstance(source_data_connections[0].location, S3Location):
        results_data_connection = DataConnection(
            connection=source_data_connections[0].connection,
            location=S3Location(bucket=source_data_connections[0].location.bucket,
                                path=output_filename)
        )
    else:
        location = DeploymentOutputAssetLocation(name=output_filename)
        results_data_connection = DataConnection(
            connection=None,
            location=location
        )
    return results_data_connection


def validate_deployment_output_connection(results_data_connection: Union['DataConnection', None],
                                          workspace: 'WorkSpace',
                                          source_data_connections: List['DataConnection'] = None) -> 'DataConnection':
    # note: if user did not provide results storage information, use default ones
    if results_data_connection is None and source_data_connections is not None:
        results_data_connection = create_results_data_connection(source_data_connections=source_data_connections,
                                                                 workspace=workspace)
    # -- end note

    if isinstance(results_data_connection.location, (AssetLocation, CP4DAssetLocation, WMLSAssetLocation, CloudAssetLocation)):
        if workspace.wml_client.default_space_id:
            results_data_connection.location.path = results_data_connection.location.path.format(option='spaces',
                                                                                                 id=workspace.wml_client.default_space_id)
        else:
            results_data_connection.location.path = results_data_connection.location.path.format(option='projects',
                                                                                                 id=workspace.wml_client.default_project_id)

    # note: results can be stored only on COS or CPD/WMLS DataAsset or DeploymentOutputAssetLocation
    if not isinstance(results_data_connection.location, (S3Location, AssetLocation, CP4DAssetLocation, WMLSAssetLocation, DeploymentOutputAssetLocation)):
        raise TypeError('Unsupported results location type. Results reference can be stored only in '
                        'one of [S3Location, AssetLocation, DeploymentOutputAssetLocation].')
    # -- end
    return results_data_connection



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
__all__ = [
    'is_run_id_exists'
]

from typing import Dict, Optional

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure


def is_run_id_exists(wml_credentials: Dict, run_id: str, space_id: Optional[str] = None) -> bool:
    """Check if specified run_id exists for WML client initialized with passed credentials.

    :param wml_credentials: WML Service Instance credentials
    :type wml_credentials: dict

    :param run_id: training run id of AutoAI experiment
    :type run_id: str

    :param space_id: optional space id for WMLS and CP4D
    :type space_id: str, optional
    """
    client = APIClient(wml_credentials)

    if space_id is not None:
        client.set.default_space(space_id)

    try:
        client.training.get_details(run_id, _internal=True)

    except ApiRequestFailure as e:
        if 'Status code: 404' in str(e):
            return False

        else:
            raise e

    return True



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact import RuntimesArtifact, RuntimesArtifactReader, RuntimesArtifactLoader


logger = logging.getLogger('WmlRuntimesAdapter')


class WmlRuntimesAdapter(object):
    """
    Adapter creating runtimes artifact using output from service.
    """

    def __init__(self, runtimes_output, client):

        self.runtimes_output = runtimes_output
        self.client = client
        self.runtimes_entity = runtimes_output.entity
        self.runtimes_metadata = runtimes_output.metadata
        self.uid = runtimes_output.metadata.guid
        self.name = None

    def artifact(self):
        runtimes_artifact_builder = type(
            "RuntimesArtifact",
            (RuntimesArtifactLoader, RuntimesArtifact, RuntimesArtifactReader, object),
            {}
        )

        prop_map = {
            MetaNames.CREATION_TIME: self.runtimes_metadata.created_at,
            MetaNames.LAST_UPDATED: self.runtimes_metadata.modified_at,
            MetaNames.RUNTIMES.URL: self.runtimes_metadata.url
        }

        if self.runtimes_entity.get('name') is not None:
            prop_map[MetaNames.RUNTIMES.NAME] = self.runtimes_entity.get('name')
            self.name = self.runtimes_entity.get('name')

        if self.runtimes_entity.get('description', None) is not None:
            prop_map[MetaNames.RUNTIMES.DESCRIPTION] = self.runtimes_entity.get('description',None)

        if self.runtimes_entity.get('platform') is not None:
            prop_map[MetaNames.RUNTIMES.PLATFORM] = self.runtimes_entity.get('platform')

        if self.runtimes_entity.get('custom_libraries') is not None:
            prop_map[MetaNames.RUNTIMES.CUSTOM_LIBRARIES_URLS] = self.runtimes_entity.get('custom_libraries')

        runtimes_content_url = self.runtimes_metadata.url + "/content"

        prop_map[MetaNames.RUNTIMES.CONTENT_URL] = runtimes_content_url

        local_meta = MetaProps(prop_map)

        runtimes_artifact = runtimes_artifact_builder(
            uid=self.uid,
            name=self.name,
            meta_props=local_meta
        )

        runtimes_artifact.client = self.client
        runtimes_artifact._content_href = runtimes_content_url

        return runtimes_artifact



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
This module will host all the exceptions which are raised by fl components
"""


class FLException(Exception):
    pass


class DuplicateRouteException(FLException):
    pass


class InvalidConfigurationException(FLException):
    pass


class InvalidServerConfigurationException(InvalidConfigurationException):
    pass


class NotFoundException(FLException):
    pass


class LocalTrainingException(FLException):
    pass


class TCPMessageOutOfOrder(FLException):
    pass


class GlobalTrainingException(FLException):
    pass


class ModelException(FLException):
    pass


class ModelInitializationException(FLException):
    pass


class ModelUpdateException(FLException):
    pass


class HyperparamsException(FLException):
    pass


class CryptoKeyGenerationException(FLException):
    pass


class CryptoException(FLException):
    pass


class AuthorityException(FLException):
    pass


class KeyManagerException(FLException):
    pass


class WarmStartException(FLException):
    pass


class FusionException(FLException):
    pass


class QuorumException(FLException):
    pass



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
"""
 An enumeration class for the crypto type field which describe what
 kind of data is being sent inside the Message
"""
from enum import Enum


class CryptoEnum(Enum):
    """
    Crypto types used for secure aggregation
    """
    CRYPTO_PAILLIER = 'Paillier'
    CRYPTO_THRESHOLD_PAILLIER = 'ThresholdPaillier'
    CRYPTO_MIFE = 'MIFE'
    CRYPTO_MCFE = 'MCFE'
    CRYPTO_DECENTRALIZED_MIFE = 'DMIFE'
    CRYPTO_FHE = 'FHE'

    KEY_PUBLIC_PARAMETER = 'pp'
    KEY_PRIVATE = 'sk'
    KEY_DECTYPT = 'dk'
    WEIGHTS_PLAINTEXT = 'plaintext-weights'
    WEIGHTS_CIPHERTEXT = 'ciphertext-weights'



"""
    IBM Watson Machine Learning REST API

    ## Authorization  ### IBM Watson Machine Learning Credentials  To start working with API one needs to generate an `access token` using the `username` and `password` available on the Service Credentials tab of the IBM Watson Machine Learning service instance or also available in the VCAP environment variable.  Example of the Service Credentials:  ```json {     \"url\": \"https://ibm-watson-ml.mybluemix.net\",     \"access_key\": \"ERY9vcBfE4sE+F4g8hcotF9L+j81WXWeZv\",     \"username\": \"c1ef4b80-2ee2-458e-ab92-e9ca97ec657d\",     \"password\": \"030528d4-5a3e-4d4c-9258-5d553513be6f\" } ```  Example of obtaining `access token` from Token Endpoint using HTTP Basic Auth (for details please refer to Token section below):  ` curl --basic --user username:password https://ibm-watson-ml.mybluemix.net/v3/identity/token `   ### Apache Spark Service Credentials  The IBM Watson Machine Learning co-operates with the Apache Spark as a Service to deploy pipeline models. For API methods requiring Apache Spark Service instance a custom header: `X-Spark-Service-Instance` with Service Credentials must be specified. The header value is a **base64 encoded** string with the Service Credentials JSON data.  [Example of API method requiring Apache Spark Service](https://console.ng.bluemix.net/docs/services/PredictiveModeling/index-gentopic1.html#pm_service_api_spark_batch)

    OpenAPI spec version: 2.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

"""

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import absolute_import

# import models into sdk package
from .models.array_model_metrics_output import ArrayModelMetricsOutput
from .models.array_model_output import ArrayModelOutput
from .models.array_model_version_output import ArrayModelVersionOutput
from .models.array_pipeline_output import ArrayPipelineOutput
from .models.array_pipeline_version_output import ArrayPipelineVersionOutput
from .models.artifact_author import ArtifactAuthor
from .models.artifact_version_metadata import ArtifactVersionMetadata
from .models.artifact_version_short_metadata import ArtifactVersionShortMetadata
from .models.batch_deploy_output import BatchDeployOutput
from .models.batch_deploy_output_entity import BatchDeployOutputEntity
from .models.batch_deploy_output_entity_execution import BatchDeployOutputEntityExecution
from .models.batch_deploy_output_meta import BatchDeployOutputMeta
from .models.batch_input import BatchInput
from .models.batch_output import BatchOutput
from .models.batch_output_array import BatchOutputArray
from .models.cols_output import ColsOutput
from .models.connection import Connection
from .models.deploy_input import DeployInput
from .models.error_message import ErrorMessage
from .models.evaluation_definition import EvaluationDefinition
from .models.evaluation_definition_metrics import EvaluationDefinitionMetrics
from .models.input_data_schema import InputDataSchema
from .models.internal_input_batch import InternalInputBatch
from .models.internal_output_batch import InternalOutputBatch
from .models.json_patch_array import JsonPatchArray
from .models.json_patch_entity import JsonPatchEntity
from .models.meta_object import MetaObject
from .models.meta_object_metadata import MetaObjectMetadata
from .models.model_input import ModelInput
from .models.model_metrics import ModelMetrics
from .models.model_metrics_values import ModelMetricsValues
from .models.model_output import ModelOutput
from .models.model_output_entity import ModelOutputEntity
from .models.model_output_entity_pipeline_version import ModelOutputEntityPipelineVersion
from .models.model_training_data_ref import ModelTrainingDataRef
from .models.model_type import ModelType
from .models.model_version_input import ModelVersionInput
from .models.model_version_output import ModelVersionOutput
from .models.model_version_output_entity import ModelVersionOutputEntity
from .models.model_version_output_entity_model import ModelVersionOutputEntityModel
from .models.online_deploy_output import OnlineDeployOutput
from .models.online_output import OnlineOutput
from .models.online_output_array import OnlineOutputArray
from .models.pipeline_input import PipelineInput
from .models.pipeline_output import PipelineOutput
from .models.pipeline_output_entity import PipelineOutputEntity
from .models.pipeline_type import PipelineType
from .models.pipeline_version_input import PipelineVersionInput
from .models.pipeline_version_output import PipelineVersionOutput
from .models.pipeline_version_output_entity import PipelineVersionOutputEntity
from .models.pipeline_version_output_entity_parent import PipelineVersionOutputEntityParent
from .models.runtime_environment import RuntimeEnvironment
from .models.score_input import ScoreInput
from .models.score_output import ScoreOutput
from .models.spark_service import SparkService
from .models.stream_input_internal import StreamInputInternal
from .models.stream_internal import StreamInternal
from .models.stream_output import StreamOutput
from .models.stream_output_array import StreamOutputArray
from .models.stream_output_internal import StreamOutputInternal
from .models.token_response import TokenResponse
from .models.training_data_schema import TrainingDataSchema
from .models.output_data_schema import OutputDataSchema
from .models.model_content_location import ModelContentLocation
from .models.training_models import TrainingModels
from .models.space_models import SpaceModels
from .models.runtime_models import RuntimeModels
from .models.software_spec_models import SoftwareSpecModels
from .models.model_schemas import ModelSchemas
from .models.pipeline_models import PipelinesModels
from .models.model_content_location import ModelContentLocation
from .models.custom_models import ModelsCustom
from .models.metrics_models import ModelsMetrics
from .models.model_schemas import ModelSchemas
from .models.size_models import ModelsSize
from .models.training_models import TrainingModels
from .models.model_definition_models import ModelDefinitionModels




# import v3 models into sdk package
from .models.array_data_input_repository import ArrayDataInputRepository
from .models.artifact_version_repository import ArtifactVersionRepository
from .models.author_repository import AuthorRepository
from .models.connection_object_with_name_repository import ConnectionObjectWithNameRepository
from .models.error_repository import ErrorRepository
from .models.error_repository_target import ErrorRepositoryTarget
from .models.error_schema_repository import ErrorSchemaRepository
from .models.evaluation_definition_repository import EvaluationDefinitionRepository
from .models.evaluation_definition_repository_metrics import EvaluationDefinitionRepositoryMetrics
from .models.framework_output_repository import FrameworkOutputRepository
from .models.framework_output_repository_libraries import FrameworkOutputRepositoryLibraries
from .models.framework_output_repository_runtimes import  FrameworkOutputRepositoryRuntimes
from .models.meta_object_repository import MetaObjectRepository
from .models.meta_object_repository_metadata import MetaObjectRepositoryMetadata
from .models.ml_assets_create_experiment_input import MlAssetsCreateExperimentInput
from .models.ml_assets_create_experiment_output import MlAssetsCreateExperimentOutput
from .models.ml_assets_create_experiment_output_array import MlAssetsCreateExperimentOutputArray
from .models.ml_assets_create_model_input import MlAssetsCreateModelInput
from .models.ml_assets_create_model_output import MlAssetsCreateModelOutput
from .models.ml_assets_create_model_output_array import MlAssetsCreateModelOutputArray
from .models.content_location import ContentLocation
from .models.content_status import ContentStatus
from .models.ml_assets_model_size_output import MlAssetsModelSizeOutput
from .models.ml_assets_get_presigned_url_output import MlAssetsGetPreSignedUrlOutput
from .models.ml_assets_upload_content_output import MlAssetsUploadContentOutput
from .models.ml_assets_upload_content_output_metadata import MlAssetsUploadContentOutputMetadata
from .models.ml_assets_upload_content_output_entity import MlAssetsUploadContentOutputEntity



# import models into sdk package
from .models.array_model_version_metrics_experiments import ArrayModelVersionMetricsExperiments
from .models.array_training_output_experiments import ArrayTrainingOutputExperiments
from .models.author_experiments import AuthorExperiments
from .models.compute_configuration_experiments import ComputeConfigurationExperiments
from .models.connection_object_source_experiments import ConnectionObjectSourceExperiments
from .models.connection_object_target_experiments import ConnectionObjectTargetExperiments
from .models.error_experiments import ErrorExperiments
from .models.error_experiments_target import ErrorExperimentsTarget
from .models.error_schema_experiments import ErrorSchemaExperiments
from .models.evaluation_definition_experiments import EvaluationDefinitionExperiments
from .models.tag_repository import TagRepository
from .models.experiment_input import ExperimentInput
from .models.experiment_input_settings import ExperimentInputSettings
from .models.experiment_output_array import ExperimentOutputArray
from .models.experiment_output_array_first import ExperimentOutputArrayFirst
from .models.experiment_output import ExperimentOutput
from .models.experiment_patch import ExperimentPatch
from .models.experiment_status_experiments import ExperimentStatusExperiments
from .models.hyper_parameters_experiments import HyperParametersExperiments
from .models.hyper_parameters_experiments_inner_values_range import HyperParametersExperimentsInnerValuesRange
from .models.hyper_parameters_for_status_experiments import HyperParametersForStatusExperiments
from .models.hyper_parameters_for_status_experiments_inner import HyperParametersForStatusExperimentsInner
from .models.hyper_parameters_optimization_experiments import HyperParametersOptimizationExperiments
from .models.meta_object_experiments import MetaObjectExperiments
from .models.meta_object_experiments_metadata import MetaObjectExperimentsMetadata
from .models.metric_object_experiments import MetricObjectExperiments
from .models.model_version_metrics_experiments import ModelVersionMetricsExperiments
from .models.patch_operation_experiments import PatchOperationExperiments
from .models.training_output_experiments import TrainingOutputExperiments
from .models.training_reference_experiments import TrainingReferenceExperiments
from .models.training_status_experiments import TrainingStatusExperiments
from .models.training_status_experiments_result import TrainingStatusExperimentsResult
from .models.hyper_parameters import HyperParameters
from .models.hyper_parameters_optimization_experiments_method import HyperParametersOptimizationExperimentsMethod
from .models.hyper_parameters_optimization_experiments_method_parameters import HyperParametersOptimizationExperimentsMethodParameters
from .models.hyper_parameters_experiments_double_range import HyperParametersExperimentsDoubleRange
from .models.hyper_parameters_experiments_int_range import HyperParametersExperimentsIntRange














# import apis into sdk package
from .apis.repository_api import RepositoryApi
from .apis.token_api import TokenApi

# import ApiClient
from .api_client import ApiClient

from .configuration import Configuration

configuration = Configuration()



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
import numpy as np

from ibmfl.party.training.local_training_handler import \
    LocalTrainingHandler

logger = logging.getLogger(__name__)


class PFNMLocalTrainingHandler(LocalTrainingHandler):

    def train(self,  fit_params=None):
        """
        Train locally using fl_model. At the end of training, a
        model_update with the new model information is generated and
        send through the connection.

        :param fit_params: (optional) Query instruction from aggregator
        :type fit_params: `dict`
        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """

        train_data, (_) = self.data_handler.get_data()
        y = train_data[1]

        y_unq, y_cnts = np.unique(y, return_counts=True)
        _class_counts = {
            int(cls_label): int(cls_counts) for (cls_label, cls_counts) in
            zip(y_unq, y_cnts)
        }

        self.update_model(fit_params['model_update'])

        self.get_train_metrics_pre()

        logger.info('Local training started...')

        self.fl_model.fit_model(train_data, fit_params, local_params=self.hyperparams)

        update = self.fl_model.get_model_update()
        update.add('class_counts', _class_counts)

        # Different frameworks have different return dimension.
        # PyTorch returns (features_out, features_in) dimensional weights for linear layers
        # Keras returns (features_in, features_out) dimensional weights for linear layers
        # Thus, this flag indicates if a transpose of the weight matrix is required or not

        update.add('transpose_weight', self.fl_model.model_type == 'PyTorch')
        logger.info('Local training done, generating model update...')

        self.get_train_metrics_post()

        return update



import logging
import joblib
import numpy as np
import time

from sklearn.linear_model import SGDClassifier, SGDRegressor
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError


from ibmfl.util import fl_metrics
from ibmfl.util import config
from ibmfl.model.sklearn_fl_model import SklearnFLModel
from ibmfl.model.model_update import ModelUpdate
from ibmfl.exceptions import LocalTrainingException, \
    ModelInitializationException, ModelException

logger = logging.getLogger(__name__)


class SklearnSGDFLModel(SklearnFLModel):
    """
    Wrapper class for sklearn.linear_model.SGDClassifier and
    sklearn.linear_model.SGDRegressor.
    """

    def __init__(self, model_name, model_spec, sklearn_model=None, **kwargs):
        """
        Create a `SklearnSGDFLModel` instance from a
        sklearn.linear_model.SGDClassifier or a
        sklearn.linear_model.SGDRegressor model.
        If sklearn_model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: A name specifying the type of model, e.g., \
        linear_SVM
        :type model_name: `str`
        :param model_spec: A dictionary contains model specification
        :type model_spec: `dict`
        :param sklearn_model: Compiled sklearn model
        :type sklearn_model: `sklearn.linear_model`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a sklearn.linear_model model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec,
                         sklearn_model=sklearn_model,
                         **kwargs)
        self.model_type = 'Sklearn-linear-SGD'
        if sklearn_model:
            if not issubclass(type(sklearn_model), (SGDClassifier,
                                                    SGDRegressor)):
                raise ValueError('Compiled sklearn model needs to be provided'
                                 '(sklearn.linear_model). '
                                 'Type provided: ' + str(type(sklearn_model)))

            self.model = sklearn_model
            
        if type(self.model) is SGDClassifier:
            self.is_classification = True
        elif type(self.model) is SGDRegressor:
            self.is_classification = False

        # checking if classes_ is provided for classification problem
        if isinstance(self.model, SGDClassifier) and \
                hasattr(self.model, "classes_"):
            # check the type
            if isinstance(self.model.classes_, np.ndarray) or \
                    self.model.classes_ is None:
                self.classes = self.model.classes_
            else:
                raise ValueError(
                    "Provided SGDClassifier model has wrong type of "
                    "`classes_` attribute. "
                    "`classes_` should be of type `numpy.ndarray`. "
                    "Instead it is of type " + str(type(self.model.classes_)))
        else:
            self.classes = None

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data a tuple \
        given in the form (x_train, y_train).
        :type train_data: `np.ndarray`
        :param fit_params: (Optional) Dictionary with hyperparameters that \
        will be used to call sklearn.linear_model fit function. \
        Provided hyperparameter should only contains parameters that \
        match sklearn expected values, e.g., `learning_rate`, which provides \
        the learning rate schedule. \
        If no `learning_rate` or `max_iter` is provided, a default value will \
        be used ( `optimal` and `1`, respectively).
        :return: None
        """
        # Default values
        max_iter = 1
        warm_start = True

        # Extract x_train and y_train, by default,
        # label is stored in the last column
        x = train_data[0]
        y = train_data[1]

        hyperparams = fit_params.get('hyperparams', {}) or {} if fit_params else {}
        sample_weight = fit_params.get('sample_weight', None) if fit_params else {}

        local_hp = hyperparams.get('local', {}) or {}
        training_hp = local_hp.get('training', {}) or {}
        
        try:
            
            if 'max_iter' not in training_hp:
                training_hp['max_iter'] = max_iter
                logger.info('Using default max_iter: ' + str(max_iter))

            # set warm_start to True
            training_hp['warm_start'] = warm_start
            logger.info('Set warm_start as ' + str(warm_start))

            for key, val in training_hp.items():
                self.model.set_params(**{key: val})
        except Exception as e:
            logger.exception(str(e))
            raise LocalTrainingException(
                'Error occurred while setting up model parameters')

        try:
            # Compute the classes based on labels if classification problem
            if isinstance(self.model, SGDClassifier):
                if self.classes is None:
                    logger.warning(
                        "Obtaining class labels based on local dataset. "
                        "This may cause failures during aggregation "
                        "when parties have distinctive class labels. ")
                    self.classes = self.get_classes(labels=y)
                # sklearn `partial_fit` uses max_iter = 1,
                # manually start the local training cycles
                for iter in range(training_hp['max_iter']):
                    logger.info("Local training epoch " + str(iter+1) + ":")
                    self.model.partial_fit(x, y,
                                           classes=self.classes,
                                           sample_weight=sample_weight)
            elif isinstance(self.model, SGDRegressor):
                for iter in range(training_hp['max_iter']):
                    logger.info("Local training epoch " + str(iter + 1) + ":")
                    self.model.partial_fit(x, y, sample_weight=sample_weight)
            else:
                raise NotImplementedError
        except Exception as e:
            logger.info(str(e))
            raise LocalTrainingException(
                'Error occurred while performing model.fit'
            )

    def update_model(self, model_update):
        """
        Update sklearn model with provided model_update, where model_update
        should contains `coef_` and `intercept_` having the same dimension
        as expected by the sklearn.linear_model.
        `coef_` : np.ndarray, shape (1, n_features) if n_classes == 2
        else (n_classes, n_features)
        `intercept_` : np.ndarray, shape (1,) if n_classes == 2 else (n_classes,)

        :param model_update: `ModelUpdate` object that contains the coef_ and \
        the intercept vectors that will be used to update the model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if isinstance(model_update, ModelUpdate):
            weights = model_update.get('weights')

            if isinstance(self.model, SGDClassifier):
                coef = np.array(weights)[:, :-1]
                intercept = np.array(weights)[:, -1]
            elif isinstance(self.model, SGDRegressor):
                coef = np.array(weights)[:-1]
                intercept = np.array(weights)[-1].reshape(1,)
            else:
                raise LocalTrainingException(
                    "Expecting scitkit-learn model of "
                    "type either "
                    "sklearn.linear_model.SGDClassifier "
                    "or sklearn.linear_model.SGDRegressor."
                    "Instead provided model is of type "
                    + str(type(self.model)))
            try:
                self.model.coef_ = coef
                self.model.intercept_ = intercept
            except Exception as e:
                raise LocalTrainingException('Error occurred during '
                                             'updating the model weights. ' +
                                             str(e))
        else:
            raise LocalTrainingException('Provided model_update should be of '
                                         'type ModelUpdate. '
                                         'Instead they are: ' +
                                         str(type(model_update)))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """

        coef = self.model.coef_
        intercept = self.model.intercept_
        if isinstance(self.model, SGDClassifier):
            n_classes = len(self.get_classes())
            # intercept is of shape (1,) if n_classes == 2 else (n_classes,)
            if n_classes == 2:
                intercept = np.reshape(intercept, [1, 1])
            else:
                intercept = np.reshape(intercept, [n_classes, 1])
            w = np.append(coef, intercept, axis=1)
        elif isinstance(self.model, SGDRegressor):
            w = np.append(coef, intercept)
        else:
            raise LocalTrainingException("Expecting scitkit-learn model of "
                                         "type either "
                                         "sklearn.linear_model.SGDClassifier "
                                         "or sklearn.linear_model.SGDRegressor."
                                         "Instead provided model is of type "
                                         + str(type(self.model)))

        return ModelUpdate(weights=w.tolist(),
                           coef=self.model.coef_,
                           intercept=self.model.intercept_)

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, test) or a datagenerator of of type `keras.utils.Sequence`,
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`

        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test)

        else:
            raise ModelException("Invalid test dataset!")

    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates the model given test data x and the corresponding labels y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding true labels to x
        :type y: `np.ndarray`
        :param kwargs: Optional sample weights accepted by model.score
        :return: score, mean accuracy on the given test data and labels
        :rtype: `dict`
        """
        acc = {}
        acc['score'] = self.model.score(x, y, **kwargs)

        y_pred = self.predict(x, **kwargs)

        if self.is_classification:
            acc['acc'] = acc['score']
            additional_metrics = fl_metrics.get_eval_metrics_for_classificaton(
                y, y_pred)
        else:
            additional_metrics = fl_metrics.get_eval_metrics_for_regression(
                y, y_pred)

        dict_metrics = {**acc, **additional_metrics}

        return dict_metrics

    def predict_proba(self, x):
        """
        Perform prediction for the given input.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :return: Array of predictions
        :rtype: `np.ndarray`
        """
        return self.model.predict_proba(x)

    def get_classes(self, labels=None):
        """
        Returns an array of shape (n_classes,). If self.classes is not None,
        return self.classes, else obtains the array based on provided labels.

        :param labels: Provided class labels to obtain the array of classes.
        :type labels: `numpy.ndarray`
        :return: An array of shape `(n_classes,)`.
        :rtype: `numpy.ndarray`
        """
        if isinstance(self.model, SGDRegressor):
            raise ValueError(
                "SGDRegressor is used for regression problems and "
                "does not have `classes_`.")
        if self.classes is not None:
            return self.classes
        elif hasattr(self.model, "classes_"):
            return self.model.classes_
        elif labels is not None:
            return np.unique(labels)
        else:
            raise NotFittedError(
                "The scikit-learn model has not been initialized with "
                "`classes_` attribute, "
                "please either manually specify `classes_` attribute as "
                "an array of shape (n_classes,) or "
                "provide labels to obtain the array of classes. ")

    def save_model(self, filename=None, path=None):
        """
        Save a sklearn model to file in the format specific
        to the framework requirement.

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in the default data location of \
        the library `DATA_PATH`.
        :type path: `str`
        :return: filename
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            filename = '{}_{}.pickle'.format(file, time.time())            

        full_path = super().get_model_absolute_path(filename)

        # check if the classes_ attribute exists for SGDClassifier
        # this attribute is required for prediction and scoring
        if isinstance(self.model, SGDClassifier) and \
                not hasattr(self.model, "classes_"):
            logger.warning(
                "The classification model to be saved has no `classes_` "
                "attribute and cannot be used for prediction!")

        with open(full_path, 'wb') as f:
            joblib.dump(self.model, f)
        logger.info('Model saved in path: %s.', full_path)
        return filename

    @staticmethod
    def load_model_from_spec(model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains the following items: model_spec['model_definition']
        contains the model definition as
        type sklearn.linear_model.SGDClassifier
        or sklearn.linear_model.SGDRegressor.

        :param model_spec: Model specification contains \
        a compiled sklearn model.
        :param model_spec: `dict`
        :return: model
        :rtype: `sklearn.cluster`
        """
        model = None
        try:
            if 'model_definition' in model_spec:
                model_file = model_spec['model_definition']
                model_absolute_path = config.get_absolute_path(model_file)

                with open(model_absolute_path, 'rb') as f:
                    model = joblib.load(f)

                if not issubclass(type(model), (SGDClassifier, SGDRegressor)):
                    raise ValueError('Provided compiled model in model_spec '
                                     'should be of type sklearn.linear_model.'
                                     'Instead it is: ' + str(type(model)))
        except Exception as ex:
            raise ModelInitializationException('Model specification was '
                                               'badly formed. '+ str(ex))
        return model

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not.

        :return: res
        :rtype: `bool`
        """
        try:
            check_is_fitted(self.model, ("coef_", "intercept_"))
        except NotFittedError as ex:
            logger.warning(
                "Model has no attribute `coef_` and `intercept_`, "
                "and hence is not fitted yet.")
            return False
        return True




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import get_url, INSTANCE_DETAILS_TYPE, is_python_2, inherited_docstring
from ibm_watson_machine_learning.metanames import ModelMetaNames, ExperimentMetaNames, FunctionMetaNames, PipelineMetanames, SpacesMetaNames, MemberMetaNames, FunctionNewMetaNames
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.models import Models
from ibm_watson_machine_learning.experiments import Experiments
from ibm_watson_machine_learning.functions import Functions
from ibm_watson_machine_learning.pipelines import Pipelines
from ibm_watson_machine_learning.spaces import Spaces
from multiprocessing import Pool
from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
from ibm_watson_machine_learning.href_definitions import API_VERSION, SPACES
import os
from warnings import warn

_DEFAULT_LIST_LENGTH = 50


class Repository(WMLResource):
    """Store and manage models, functions, spaces, pipelines and experiments
    using Watson Machine Learning Repository.
    
    To view ModelMetaNames, use:

    .. code-block:: python

        client.repository.ModelMetaNames.show()

    To view ExperimentMetaNames, use:

    .. code-block:: python

        client.repository.ExperimentMetaNames.show()

    To view FunctionMetaNames, use:

    .. code-block:: python

        client.repository.FunctionMetaNames.show()

    To view PipelineMetaNames, use:

    .. code-block:: python

        client.repository.PipelineMetaNames.show()

    """

    cloud_platform_spaces = False
    icp_platform_spaces = False

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        if not client.ICP and not client.WSD and not client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            Repository._validate_type(client.service_instance.details, u'instance_details', dict, True)
            Repository._validate_type_of_details(client.service_instance.details, INSTANCE_DETAILS_TYPE)
        self._ICP = client.ICP
        self._WSD = client.WSD
        self._ml_repository_client = None
        Repository.cloud_platform_spaces = client.CLOUD_PLATFORM_SPACES
        Repository.icp_platform_spaces = client.ICP_PLATFORM_SPACES

        self.ExperimentMetaNames = ExperimentMetaNames()
        if not client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            self.FunctionMetaNames = FunctionMetaNames()
        else:
            self.FunctionMetaNames = FunctionNewMetaNames()
        self.PipelineMetaNames = PipelineMetanames()
        self.SpacesMetaNames = SpacesMetaNames()
        self.ModelMetaNames = ModelMetaNames()
        self.MemberMetaNames = MemberMetaNames()

        self._refresh_repo_client() # regular token is initialized in service_instance

    def _refresh_repo_client(self):
        # If apiKey is passed in credentials then refresh repoclient with IAM token else MLToken
        self._ml_repository_client = MLRepositoryClient(self._wml_credentials[u'url'])
        if self._client.proceed is True:
            if self._client.service_instance._is_iam() is not None:
                self._ml_repository_client.authorize_with_token(self._client.wml_token)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
            else:
                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    platform_spaces = True
                else:
                    platform_spaces = False

                self._ml_repository_client.authorize_with_iamtoken(self._client.wml_token,
                                                                   self._wml_credentials[u'instance_id'],
                                                                   platform_spaces)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                # Cloud Convergence
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                   self._ml_repository_client._add_header('ML-Instance-ID', self._wml_credentials[u'instance_id'])
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
        else:
            if self._client._is_IAM():
                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    platform_spaces = True
                else:
                    platform_spaces = False

                self._ml_repository_client.authorize_with_iamtoken(self._client.wml_token,
                                                                   self._wml_credentials[u'instance_id'],
                                                                   platform_spaces)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                # Cloud Convergence
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                   self._ml_repository_client._add_header('ML-Instance-ID', self._wml_credentials[u'instance_id'])
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
            else:
                if self._ICP:
                    self._repotoken = self._client._get_icptoken()
                    self._ml_repository_token = self._repotoken.replace('Bearer', '')
                    self._ml_repository_client.authorize_with_token(self._ml_repository_token)
                else:
                    if not self._client.WSD:
                        self._ml_repository_client.authorize(self._wml_credentials[u'username'], self._wml_credentials[u'password'])
                        self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                        if self._client.project_id is not None:
                           self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)

    @inherited_docstring(Experiments.store, {'experiments.get_href': 'repository.get_experiment_href'})
    def store_experiment(self, meta_props):
        if self._client.WSD:
            raise WMLClientError(u'Experiment APIs are not supported in Watson Studio Desktop.')

        return self._client.experiments.store(meta_props)

    @inherited_docstring(Spaces.store)
    def store_space(self, meta_props):
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.store(meta_props)

    @inherited_docstring(Spaces.create_member)
    def create_member(self, space_uid, meta_props):
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.create_member(space_uid, meta_props)

    @staticmethod
    def _meta_props_to_repository_v3_style(meta_props):
        if is_python_2():
            new_meta_props = meta_props.copy()

            for key in new_meta_props:
                if type(new_meta_props[key]) is unicode:
                    new_meta_props[key] = str(new_meta_props[key])

            return new_meta_props
        else:
            return meta_props

    @inherited_docstring(Pipelines.store)
    def store_pipeline(self, meta_props):
        return self._client.pipelines.store(meta_props)

    @inherited_docstring(Models.store, {'store()': 'store_model()'})
    def store_model(self, model, meta_props=None, training_data=None, training_target=None, pipeline=None,
                    feature_names=None, label_column_names=None,subtrainingId=None, round_number=None,
                    experiment_metadata=None, training_id=None):
        return self._client._models.store(model, meta_props=meta_props, training_data=training_data,
                                          training_target=training_target, pipeline=pipeline,
                                          feature_names=feature_names, label_column_names=label_column_names,
                                          subtrainingId=subtrainingId, round_number=round_number,
                                          experiment_metadata=experiment_metadata, training_id=training_id)

    def clone(self, artifact_id, space_id=None, action="copy", rev_id=None):
        # """ it is not supported in v4ga
        #     Create a new resource(models, runtimes, libraries, experiments, functions, pipelines) identical with the model either in the same space or in a new space. All dependent assets will be cloned too.
        #
        #     **Parameters**
        #
        #     .. important::
        #         #. **model_id**:  Guid of the artifact to be cloned:\n
        #
        #            **type**: str\n
        #
        #         #. **space_id**: Guid of the space to which the model needs to be cloned. (optional)
        #
        #            **type**: str\n
        #
        #         #. **action**: Action specifying "copy" or "move". (optional)
        #
        #            **type**: str\n
        #
        #         #. **rev_id**: Revision ID of the artifact. (optional)
        #
        #            **type**: str\n
        #
        #     **Output**
        #
        #     .. important::
        #
        #             **returns**: Metadata of the model cloned.\n
        #             **return type**: dict\n
        #
        #     **Example**
        #
        #      >>> client.repository.clone(artifact_id=artifact_id,space_id=space_uid,action="copy")
        #
        #     .. note::
        #         * If revision id is not specified, all revisions of the artifact are cloned\n
        #
        #         * Default value of the parameter action is copy\n
        #
        #         * Space guid is mandatory for move action\n
        #
        #     """
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Cloning is not supported.')

        Models._validate_type(artifact_id, 'artifact_id', str, True)
        clone_meta = {}
        if space_id is not None:
            clone_meta["space"] = {"href": API_VERSION + SPACES + "/" + space_id}
        if action is not None:
            clone_meta["action"] = action
        if rev_id is not None:
            clone_meta["rev"] = rev_id
        res = self._check_artifact_type(artifact_id)

        url = ""
        type = ""
        if res['model'] is True:
            url = self._client.service_instance._href_definitions.get_published_model_href(artifact_id)
            type = "model"
        elif res['library'] is True:
            url = self._client.service_instance._href_definitions.get_custom_library_href(artifact_id)
            type = "library"
        elif res['runtime'] is True:
            url = self._client.service_instance._href_definitions.get_runtime_href(artifact_id)
            type = "runtime"
        elif res['function'] is True:
            url = self._client.service_instance._href_definitions.get_function_href(artifact_id)
            type = "function"
        elif res['pipeline'] is True:
            url = self._client.service_instance._href_definitions.get_pipeline_href(artifact_id)
            type = "pipeline"
        elif res['experiment'] is True:
            url = self._client.service_instance._href_definitions.get_experiment_href(artifact_id)
            type = "experiment"


        if type == "":
            raise WMLClientError('Unsupported artifact type. Supported artifact types are models, libraries, runtimes, experiments, pipelines and functions')

        response_post = requests.post(url, json=clone_meta,
                                      headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=200, operationName=u'cloning '+ type,
                                        response=response_post)

        return details

    @inherited_docstring(Functions.store)
    def store_function(self, function, meta_props):
        return self._client._functions.store(function, meta_props)

    @inherited_docstring(Models.create_revision)
    def create_model_revision(self, model_uid):
        return self._client._models.create_revision(model_uid=model_uid)

    @inherited_docstring(Pipelines.create_revision)
    def create_pipeline_revision(self, pipeline_uid):
        return self._client.pipelines.create_revision(pipeline_uid=pipeline_uid)

    @inherited_docstring(Functions.create_revision)
    def create_function_revision(self, function_uid):
        return self._client._functions.create_revision(function_uid=function_uid)

    @inherited_docstring(Experiments.create_revision, {'experiment_id': 'experiment_uid'})
    def create_experiment_revision(self, experiment_uid):
        return self._client.experiments.create_revision(experiment_id=experiment_uid)

    @inherited_docstring(Models.update, {'meta_props': 'updated_meta_props'})
    def update_model(self, model_uid, updated_meta_props=None, update_model=None):
        return self._client._models.update(model_uid, updated_meta_props, update_model)

    @inherited_docstring(Experiments.update)
    def update_experiment(self, experiment_uid, changes):
        if self._client.WSD:
            raise WMLClientError('Experiments APIs are not supported in IBM Watson Studio Desktop.')

        return self._client.experiments.update(experiment_uid, changes)

    @inherited_docstring(Functions.update)
    def update_function(self, function_uid, changes, update_function=None):
        return self._client._functions.update(function_uid, changes, update_function)

    @inherited_docstring(Pipelines.update)
    def update_pipeline(self, pipeline_uid, changes):
        return self._client.pipelines.update(pipeline_uid, changes)

    @inherited_docstring(Spaces.update)
    def update_space(self, space_uid, changes):
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.update(space_uid, changes)

    @inherited_docstring(Models.load)
    def load(self, artifact_uid):
        return self._client._models.load(artifact_uid)

    def download(self, artifact_uid, filename='downloaded_artifact.tar.gz', rev_uid=None, format=None):
        """Downloads configuration file for artifact with specified uid.

        :param artifact_uid: Unique Id of model, function, runtime or library
        :type artifact_uid: str
        :param filename: name of the file to which the artifact content has to be downloaded
        :type filename: str, optional

        :return: path to the downloaded artifact content
        :rtype: str

        **Examples**

        .. code-block:: python

            client.repository.download(model_uid, 'my_model.tar.gz')
            client.repository.download(model_uid, 'my_model.json') # if original model was saved as json, works only for xgboost 1.3
        """
        self._validate_type(artifact_uid, 'artifact_uid', str, True)
        self._validate_type(filename, 'filename', str, True)

        res = self._check_artifact_type(artifact_uid)

        if res['model'] is True:
            return self._client._models.download(artifact_uid, filename, rev_uid,format)
        elif res['function'] is True:
            return self._client._functions.download(artifact_uid, filename, rev_uid)
        elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_35 and not self._client.ICP_40  \
                and not self._client.ICP_45 and not self._client.ICP_46 and res['library'] is True:
            return self._client.runtimes.download_library(artifact_uid, filename)
        elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_35 and not self._client.ICP_40 \
                and not self._client.ICP_45 and not self._client.ICP_46 and res['runtime'] is True:
            return self._client.runtimes.download_configuration(artifact_uid, filename)
        else:
            raise WMLClientError('Unexpected type of artifact to download or Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid) )

    def delete(self, artifact_uid):
        """Delete model, experiment, pipeline, space, runtime, library or function from repository.

        :param artifact_uid: Unique id of stored model, experiment, function, pipeline, space, library or runtime
        :type artifact_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.repository.delete(artifact_uid)
        """
        Repository._validate_type(artifact_uid, u'artifact_uid', str, True)
        if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and self._if_deployment_exist_for_asset(artifact_uid):
            raise WMLClientError(
                u'Cannot delete artifact that has existing deployments. Please delete all associated deployments and try again')
        params = self._client._params()
        if Repository.cloud_platform_spaces or self._client.ICP_PLATFORM_SPACES:
            # ideally purge_on_delete=true query param has to be provided for deletion of cams assets
            # This doesn't seem to be done for CP4D 3.0.1 and before. We should do this for CP4D 3.5
            params.update({'purge_on_delete': 'true'})

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(artifact_uid),
                                       params=params,
                                       headers=self._client._get_headers())

        if response.status_code == 200 or response.status_code == 204:
            if response.status_code == 200:
                response = self._handle_response(200, u'delete assets', response)
                return response
            else:
                response = self._handle_response(204, u'delete assets', response)
                return response
        else:
            if Repository.cloud_platform_spaces or self._client.ICP_PLATFORM_SPACES:
                # Since we are using /v2/assets for deletion, don't need all the logic
                # in the following else block. The else block is applicable only for cloud beta
                # and has to be kept till then. For 3.5, move logic to same as cloud convergence
                # for deletion
                if response.status_code == 404:
                    raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))
                else:
                    raise WMLClientError("Deletion error for the given id : ", response.text)
            else:
                artifact_type = self._check_artifact_type(artifact_uid)
                self._logger.debug(u'Attempting deletion of artifact with type: \'{}\''.format(str(artifact_type)))
                if self._client.WSD:
                    if artifact_type[u'model'] is True:
                        return self._client._models.delete(artifact_uid)
                    elif artifact_type[u'pipeline'] is True:
                        return self._client.pipelines.delete(artifact_uid)
                    elif artifact_type[u'function'] is True:
                        return self._client._functions.delete(artifact_uid)
                    else:
                        raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))
                else:
                    if artifact_type[u'model'] is True:
                        return self._client._models.delete(artifact_uid)
                    elif artifact_type[u'experiment'] is True:
                        return self._client.experiments.delete(artifact_uid)
                    elif artifact_type[u'pipeline'] is True:
                        return self._client.pipelines.delete(artifact_uid)
                    elif artifact_type[u'function'] is True:
                        return self._client._functions.delete(artifact_uid)
                    elif artifact_type[u'space'] is True:
                        return self._client.spaces.delete(artifact_uid)
                    elif artifact_type[u'runtime'] is True:
                        return self._client.runtimes.delete(artifact_uid)
                    elif artifact_type[u'library'] is True:
                        return self._client.runtimes.delete_library(artifact_uid)
                    else:
                        raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))

    def get_details(self, artifact_uid=None, spec_state=None):
        """Get metadata of stored artifacts. If `artifact_uid` is not specified returns all models, experiments,
        functions, pipelines, spaces, libraries and runtimes metadata.

        :param artifact_uid: Unique Id of stored model, experiment, function, pipeline, space, library or runtime
        :type artifact_uid: str, optional

        :param spec_state: software specification state, can be used only when `artifact_uid` is None
        :type spec_state: SpecStates, optional

        :return: stored artifact(s) metadata
        :rtype: dict (if artifact_uid is not None) or {"resources": [dict]} (if artifact_uid is None)

        **Examples**

        .. code-block:: python

            details = client.repository.get_details(artifact_uid)
            details = client.repository.get_details()

        Example of getting all repository assets with deprecated software specifications:

        .. code-block:: python

            from ibm_watson_machine_learning.lifecycle import SpecStates

            details = client.repository.get_details(spec_state=SpecStates.DEPRECATED)
        """
        Repository._validate_type(artifact_uid, u'artifact_uid', str, False)

        if artifact_uid is None and self._client.WSD is None:
            model_details = self._client._models.get_details(spec_state=spec_state)
            experiment_details = self.get_experiment_details() if not spec_state else {'resources': []}
            pipeline_details = self.get_pipeline_details() if not spec_state else {'resources': []}
            function_details = self._client._functions.get_details(spec_state=spec_state)

            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                space_details = self._client.spaces.get_details() if not spec_state else {'resources': []}
                library_details = self._client.runtimes.get_library_details() if not spec_state else {'resources': []}
                runtime_details = self._client.runtimes.get_details() if not spec_state else {'resources': []}
                details = {
                    u'models': model_details,
                    u'experiments': experiment_details,
                    u'pipeline': pipeline_details,
                    u'runtimes': runtime_details,
                    u'libraries': library_details,
                    u'spaces': space_details,
                    u'functions': function_details
                }
            else:
                details = {
                    u'models': model_details,
                    u'experiments': experiment_details,
                    u'pipeline': pipeline_details,
                    u'functions': function_details
                }
        else:
            if self._client.WSD and artifact_uid is None:
                raise WMLClientError(
                        u' artifiact_uid is mandatory for get_details() in IBM Watson Studio Desktop.')
            uid_type = self._check_artifact_type(artifact_uid)
            if self._client.WSD:
                if uid_type[u'model'] is True:
                    details = self._client._models.get_details(artifact_uid)
                elif uid_type[u'pipeline'] is True:
                    details = self.get_pipeline_details(artifact_uid)
                elif uid_type[u'function'] is True:
                    details = self._client._functions.get_details(artifact_uid)
                else:
                    raise WMLClientError(
                        u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))
            else:
                if uid_type[u'model'] is True:
                    details = self._client._models.get_details(artifact_uid)
                elif uid_type[u'experiment'] is True:
                    details = self.get_experiment_details(artifact_uid)
                elif uid_type[u'pipeline'] is True:
                    details = self.get_pipeline_details(artifact_uid)
                elif uid_type[u'function'] is True:
                    details = self._client._functions.get_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'runtime'] is True:
                    details = self._client.runtimes.get_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'library'] is True:
                    details = self._client.runtimes.get_library_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'space'] is True:
                    details = self._client.spaces.get_details(artifact_uid)
                else:
                    raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))

        return details

    @inherited_docstring(Models.get_details)
    def get_model_details(self, model_uid=None, limit=None, asynchronous=False, get_all=False, spec_state=None):
        return self._client._models.get_details(model_uid, limit,
                                                asynchronous=asynchronous, get_all=get_all, spec_state=spec_state)

    @inherited_docstring(Models.get_revision_details)
    def get_model_revision_details(self, model_uid, rev_uid):
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        return self._client._models.get_revision_details(model_uid, rev_uid)

    @inherited_docstring(Experiments.get_details)
    def get_experiment_details(self, experiment_uid=None, limit=None, asynchronous=False, get_all=False):
        if self._client.WSD:
            raise WMLClientError('Experiment APIs are not supported in IBM Watson Studio Desktop.')

        Repository._validate_type(experiment_uid, u'experiment_uid', str, False)
        Repository._validate_type(limit, u'limit', int, False)
        Repository._validate_type(asynchronous, u'asynchronous', bool, False)
        Repository._validate_type(get_all, u'get_all', bool, False)

        return self._client.experiments.get_details(experiment_uid, limit, asynchronous, get_all)

    @inherited_docstring(Experiments.get_revision_details, {'rev_uid': 'rev_id'})
    def get_experiment_revision_details(self, experiment_uid, rev_id):
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')

        return self._client.experiments.get_revision_details(experiment_uid, rev_id)

    @inherited_docstring(Functions.get_details)
    def get_function_details(self, function_uid=None, limit=None, asynchronous=False, get_all=False, spec_state=None):
        Repository._validate_type(function_uid, u'function_uid', str, False)
        Repository._validate_type(limit, u'limit', int, False)
        Repository._validate_type(asynchronous, u'asynchronous', bool, False)
        Repository._validate_type(get_all, u'get_all', bool, False)
        Repository._validate_type(spec_state, u'spec_state', object, False)
        return self._client._functions.get_details(function_uid, limit, asynchronous, get_all, spec_state)

    @inherited_docstring(Functions.get_revision_details, {'rev_uid': 'rev_id'})
    def get_function_revision_details(self, function_uid, rev_id):
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Not supported in this release')
        return self._client._functions.get_revision_details(function_uid, rev_id)

    @inherited_docstring(Pipelines.get_details)
    def get_pipeline_details(self, pipeline_uid=None, limit=None, asynchronous=False, get_all=False):
        Repository._validate_type(pipeline_uid, u'pipeline_uid', str, False)
        Repository._validate_type(limit, u'limit', int, False)
        Repository._validate_type(asynchronous, u'asynchronous', bool, False)
        Repository._validate_type(get_all, u'get_all', bool, False)
        return self._client.pipelines.get_details(pipeline_uid, limit, asynchronous, get_all)

    @inherited_docstring(Pipelines.get_revision_details, {'rev_uid': 'rev_id'})
    def get_pipeline_revision_details(self, pipeline_uid, rev_id):
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        return self._client.pipelines.get_revision_details(pipeline_uid, rev_id)

    @inherited_docstring(Spaces.get_details)
    def get_space_details(self, space_uid=None, limit=None):
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        Repository._validate_type(space_uid, u'space_uid', str, False)
        Repository._validate_type(limit, u'limit', int, False)
        return self._client.spaces.get_details(space_uid, limit)

    @inherited_docstring(Spaces.get_members_details)
    def get_members_details(self, space_uid, member_id=None, limit=None):
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.get_members_details(space_uid, member_id, limit)

    @staticmethod
    @inherited_docstring(Models.get_href)
    def get_model_href(model_details):
        return Models.get_href(model_details)

    @staticmethod
    def get_model_uid(model_details):
        """
            This method is deprecated, please use ``get_id()`` instead."
        """
        warn("This method is deprecated, please use get_model_id()")
        print("This method is deprecated, please use get_model_id()")

        return Models.get_id(model_details)

    @staticmethod
    @inherited_docstring(Models.get_id)
    def get_model_id(model_details):
        return Models.get_id(model_details)

    @staticmethod
    @inherited_docstring(Experiments.get_uid, {'experiments.get_details': 'repository.get_experiment_details'})
    def get_experiment_uid(experiment_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_uid(experiment_details)

    @staticmethod
    @inherited_docstring(Experiments.get_id, {'experiments.get_details': 'repository.get_experiment_details'})
    def get_experiment_id(experiment_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_id(experiment_details)

    @staticmethod
    @inherited_docstring(Experiments.get_href, {'experiments.get_details': 'repository.get_experiment_details'})
    def get_experiment_href(experiment_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_href(experiment_details)

    @staticmethod
    @inherited_docstring(Functions.get_id)
    def get_function_id(function_details):
        return Functions.get_id(function_details)

    @staticmethod
    @inherited_docstring(Functions.get_uid)
    def get_function_uid(function_details):
        return Functions.get_uid(function_details)

    @staticmethod
    @inherited_docstring(Pipelines.get_uid)
    def get_pipeline_uid(pipeline_details):
        return Pipelines.get_uid(pipeline_details)

    @staticmethod
    @inherited_docstring(Functions.get_href)
    def get_function_href(function_details):
        return Functions.get_href(function_details)

    @staticmethod
    @inherited_docstring(Pipelines.get_href, {'pipelines.get_details': 'repository.get_pipeline_details'})
    def get_pipeline_href(pipeline_details):
        return Pipelines.get_href(pipeline_details)

    @staticmethod
    @inherited_docstring(Pipelines.get_id)
    def get_pipeline_id(pipeline_details):
        return Pipelines.get_id(pipeline_details)

    @staticmethod
    @inherited_docstring(Spaces.get_uid, {'spaces.get_details': 'repository.get_space_details'})
    def get_space_uid(space_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_uid(space_details)

    @staticmethod
    @inherited_docstring(Spaces.get_member_uid, {'spaces.get_member_details': 'repository.get_member_details'})
    def get_member_uid(member_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_member_uid(member_details)

    @staticmethod
    @inherited_docstring(Spaces.get_href, {'spaces.get_details': 'repository.get_space_details'})
    def get_space_href(space_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_href(space_details)

    @staticmethod
    @inherited_docstring(Spaces.get_member_href, {'spaces.get_member_details': 'repository.get_member_details'})
    def get_member_href(member_details):
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_member_href(member_details)

    def list(self):
        """Print stored models, pipelines, runtimes, libraries, functions, spaces and experiments in a table format.
        If limit is set to None there will be only first 50 records shown.

        **Example**
        
        .. code-block:: python

            client.repository.list()
        """

        from tabulate import tabulate

        headers = self._client._get_headers()
        params = self._client._params()
        params.update({u'limit': 1000})
        #params = {u'limit': 1000} # TODO - should be unlimited, if results not sorted

        pool = Pool(processes=4)
        isIcp = self._ICP
        if self._client.WSD:
            raise WMLClientError(
                u'list() - Listing all artifact is not supported for IBM Watson Studio Desktop. '
                u'Use list method of specific artifact.')


        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            endpoints = {
                u'model': self._client.service_instance._href_definitions.get_published_models_href(),
                u'experiment': self._client.service_instance._href_definitions.get_experiments_href(),
                u'pipeline': self._client.service_instance._href_definitions.get_pipelines_href(),
                u'function': self._client.service_instance._href_definitions.get_functions_href()
            }
        else:
            endpoints = {
                u'model': self._client.service_instance._href_definitions.get_published_models_href(),
                u'experiment': self._client.service_instance._href_definitions.get_experiments_href(),
                u'pipeline': self._client.service_instance._href_definitions.get_pipelines_href(),
                u'function': self._client.service_instance._href_definitions.get_functions_href(),
                u'runtime': self._client.service_instance._href_definitions.get_runtimes_href(),
                u'library': self._client.service_instance._href_definitions.get_custom_libraries_href()
            }

        artifact_get = {}
        for artifact in endpoints:
            if (artifact=="library" or artifact=="runtime" or artifact=="space"):
                params = None
            else:
                params = self._client._params()
            artifact_get[artifact] = pool.apply_async(get_url,
                                                (endpoints[artifact], self._client._get_headers(), params, isIcp))

        # artifact_get = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, self._client._params(), isIcp)) for
        #           artifact in endpoints if (artifact != "library" or artifact != "runtime" or artifact != "space")}
        # artifact_no_space = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, None, isIcp)) for artifact
        #                    in endpoints if (artifact == "library" or artifact == "runtime")}
        # artifact_get.update(artifact_no_space)

        resources = {artifact: [] for artifact in endpoints}

        for artifact in endpoints:
            try:
                    response = artifact_get[artifact].get()
                    response_text = self._handle_response(200, u'getting all {}s'.format(artifact), response)
                    resources[artifact] = response_text[u'resources']
            except Exception as e:
                    self._logger.error(e)

        pool.close()

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            sw_spec_info = {s['id']: s
                            for s in self._client.software_specifications.get_details(state_info=True)['resources']}

            def get_spec_info(spec_id, prop):
                if spec_id and spec_id in sw_spec_info:
                    return sw_spec_info[spec_id].get(prop, '')
                else:
                    return ''

            values = []
            for t in endpoints.keys():
                values += [
                    (m['metadata']['id'],
                     m['metadata']['name'],
                     m['metadata']['created_at'],
                     m['entity']['type'] if t == 'model' else '-',
                     t if t != 'function' else m['entity']['type'] + ' function',
                     get_spec_info(m['entity'].get('software_spec', {}).get('id'), 'state'),
                     get_spec_info(m['entity'].get('software_spec', {}).get('id'), 'replacement'))
                    for m in resources[t]]

            values = sorted(sorted(list(set(values)), key=lambda x: x[2], reverse=True), key=lambda x: x[4])

            table = tabulate([['GUID', 'NAME', 'CREATED', 'FRAMEWORK', 'TYPE', 'SPEC_STATE', 'SPEC_REPLACEMENT']]
                             + values[:_DEFAULT_LIST_LENGTH])
        else:
            model_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], m[u'entity'][u'type'], u'model') for m in resources[u'model']]
            experiment_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'],m['metadata']['created_at'], u'-', u'experiment') for m in resources[u'experiment']]
            pipeline_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', u'pipeline')for m in self._client.pipelines.get_details()[u'resources']]
            function_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'type'] + u' function') for m in resources[u'function']]
            runtime_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'platform'][u'name'] + u' runtime') for m in resources[u'runtime']]
            library_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'platform'][u'name'] + u' library') for m in resources[u'library']]
            values = list(set(model_values + experiment_values + pipeline_values + function_values + runtime_values + library_values))
            values = sorted(sorted(values, key=lambda x: x[2], reverse=True), key=lambda x: x[4])

            table = tabulate([[u'GUID', u'NAME', u'CREATED', u'FRAMEWORK', u'TYPE']] + values[:_DEFAULT_LIST_LENGTH])

        print(table)
        if len(values) > _DEFAULT_LIST_LENGTH:
             print('Note: Only first {} records were displayed. To display more use more specific list functions.'.format(_DEFAULT_LIST_LENGTH))

    @inherited_docstring(Models.list)
    def list_models(self, limit=None, asynchronous=False, get_all=False):
        return self._client._models.list(limit=limit, asynchronous=asynchronous, get_all=get_all)

    @inherited_docstring(Experiments.list)
    def list_experiments(self, limit=None):
        if self._client.WSD:
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        self._client.experiments.list(limit=limit)

    @inherited_docstring(Spaces.list)
    def list_spaces(self, limit=None):
        if self._client.WSD:
            raise WMLClientError('list_spaces - Listing spaces is not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        self._client.spaces.list(limit=limit)

    @inherited_docstring(Functions.list)
    def list_functions(self, limit=None):
        self._client._functions.list(limit=limit)

    @inherited_docstring(Pipelines.list)
    def list_pipelines(self, limit=None):
        self._client.pipelines.list(limit=limit)

    @inherited_docstring(Spaces.list_members)
    def list_members(self, space_uid, limit=None):
        if self._client.WSD:
            raise WMLClientError('list_members - Listing members is not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        self._client.spaces.list_members(space_uid=space_uid, limit=limit)

    def _check_artifact_type(self, artifact_uid):
        Repository._validate_type(artifact_uid, u'artifact_uid', str, True)

        def _artifact_exists(response):
            return (response is not None) and (u'status_code' in dir(response)) and (response.status_code == 200)

        pool = Pool(processes=4)
        #headers =

        isIcp=self._ICP

        if self._client.WSD:
            endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + artifact_uid
            response = requests.get(
                endpoint,
                params=self._client._params()
            )

           # requestsget_url, (endpoint, self._client._get_headers(), self._client._params(), True))
            response_get = _artifact_exists(response)

            artifact_type = artifact_uid.rsplit(".")[0]
            artifact_list = ['wml_model', 'wml_pipeline', 'wml_function']
            artifact_type_exists = {artifact.rsplit('_')[-1]: (response_get and artifact == artifact_type) for artifact in artifact_list}
            return artifact_type_exists

        else:
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                endpoints = {
                    u'model': self._client.service_instance._href_definitions.get_model_last_version_href(artifact_uid),
                    u'pipeline': self._client.service_instance._href_definitions.get_pipeline_href(artifact_uid),
                    u'experiment': self._client.service_instance._href_definitions.get_experiment_href(artifact_uid),
                    u'function': self._client.service_instance._href_definitions.get_function_href(artifact_uid)
                }
            else:
                endpoints = {
                        u'model': self._client.service_instance._href_definitions.get_model_last_version_href(artifact_uid),
                        u'pipeline': self._client.service_instance._href_definitions.get_pipeline_href(artifact_uid),
                        u'experiment': self._client.service_instance._href_definitions.get_experiment_href(artifact_uid),
                        u'function': self._client.service_instance._href_definitions.get_function_href(artifact_uid),
                        u'runtime': self._client.service_instance._href_definitions.get_runtime_href(artifact_uid),
                        u'library': self._client.service_instance._href_definitions.get_custom_library_href(artifact_uid),
                        u'space': self._client.service_instance._href_definitions.get_space_href(artifact_uid)
                    }
            future = {}
            for artifact in endpoints:
                if (artifact=="library" or artifact=="runtime" or artifact=="space"):
                    params = None
                else:
                    params = self._client._params()
                future[artifact] = pool.apply_async(get_url, (endpoints[artifact], self._client._get_headers(), params , isIcp))

            # future_no_space = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, None, isIcp)) for artifact in endpoints if (artifact=="library" or artifact=="runtime" or artifact=="space")}
            # future.update(future_no_space)
            response_get = {artifact: None for artifact in endpoints}

            for artifact in endpoints:
                    try:
                        response_get[artifact] = future[artifact].get(timeout=180)
                        self._logger.debug(u'Response({})[{}]: {}'.format(endpoints[artifact], response_get[artifact].status_code, response_get[artifact].text))

                    except Exception as e:
                        self._logger.debug(u'Error during checking artifact type: ' + str(e))

            pool.close()
            artifact_type = {artifact: _artifact_exists(response_get[artifact]) for artifact in response_get}

            return artifact_type

    def create_revision(self, artifact_uid):
        """Create revision for passed `artifact_uid`.

        :param artifact_uid: Unique id of stored model, experiment, function or pipelines
        :type artifact_uid: str

        :return: artifact new revision metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            details = client.repository.create_revision(artifact_uid)
        """
        Repository._validate_type(artifact_uid, u'artifact_uid', str, True)

        uid_type = self._check_artifact_type(artifact_uid)
        if uid_type[u'experiment'] is True:
            return self._client.experiments.create_revision(artifact_uid)
        if uid_type[u'pipeline'] is True:
            return self._client.pipelines.create_revision(artifact_uid)
        else:
           raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))

        return details

    def _get_revision_details(self, artifact_uid):
        """Get metadata of stored artifacts revisions.

        :param artifact_uid:  unique id of stored model or experiment or function or pipelines
        :type artifact_uid: str

        :return: stored artifacts metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            details = client.repository.get_revision_details(artifact_uid)

        """
        Repository._validate_type(artifact_uid, u'artifact_uid', str, True)

        uid_type = self._check_artifact_type(artifact_uid)

        if uid_type[u'experiment'] is True:
            details = self._client.experiments.get_revision_details(artifact_uid)
        if uid_type[u'pipeline'] is True:
            details = self._client.pipelines.get_revisions(artifact_uid)

        else:
            raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))
        return details

    @inherited_docstring(Models.list_revisions)
    def list_models_revisions(self, model_uid, limit=None):
        self._client._models.list_revisions(model_uid, limit=limit)

    @inherited_docstring(Pipelines.list_revisions)
    def list_pipelines_revisions(self, pipeline_uid, limit=None):
        self._client.pipelines.list_revisions(pipeline_uid, limit=limit)

    @inherited_docstring(Functions.list_revisions)
    def list_functions_revisions(self, function_uid, limit=None):
        self._client._functions.list_revisions(function_uid, limit=limit)

    @inherited_docstring(Experiments.list_revisions)
    def list_experiments_revisions(self, experiment_uid, limit=None):
        self._client.experiments.list_revisions(experiment_uid, limit=limit)

    @inherited_docstring(Models.promote)
    def promote_model(self, model_id: str, source_project_id: str, target_space_id: str):  # deprecated
        return self._client._models.promote(model_id, source_project_id, target_space_id)





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import re

TRAINING_MODEL_HREF_PATTERN = u'{}/v4/trainings/{}'
TRAINING_MODELS_HREF_PATTERN = u'{}/v4/trainings'
REPO_MODELS_FRAMEWORKS_HREF_PATTERN = u'{}/v3/models/frameworks'

INSTANCE_ENDPOINT_HREF_PATTERN = u'{}/v3/wml_instance'
INSTANCE_BY_ID_ENDPOINT_HREF_PATTERN = u'{}/v3/wml_instances/{}'
TOKEN_ENDPOINT_HREF_PATTERN = u'{}/v3/identity/token'
CPD_TOKEN_ENDPOINT_HREF_PATTERN = u'{}/icp4d-api/v1/authorize'
CPD_BEDROCK_TOKEN_ENDPOINT_HREF_PATTERN = u'{}/idprovider/v1/auth/identitytoken'
CPD_VALIDATION_TOKEN_ENDPOINT_HREF_PATTERN = u'{}/v1/preauth/validateAuth'
EXPERIMENTS_HREF_PATTERN = u'{}/v4/experiments'
EXPERIMENT_HREF_PATTERN = u'{}/v4/experiments/{}'
EXPERIMENT_RUNS_HREF_PATTERN = u'{}/v3/experiments/{}/runs'
EXPERIMENT_RUN_HREF_PATTERN = u'{}/v3/experiments/{}/runs/{}'

PUBLISHED_MODEL_HREF_PATTERN = u'{}/v4/models/{}'
PUBLISHED_MODELS_HREF_PATTERN = u'{}/v4/models'
LEARNING_CONFIGURATION_HREF_PATTERN = u'{}/v3/wml_instances/{}/published_models/{}/learning_configuration'
LEARNING_ITERATION_HREF_PATTERN = u'{}/v3/wml_instances/{}/published_models/{}/learning_iterations/{}'
LEARNING_ITERATIONS_HREF_PATTERN = u'{}/v3/wml_instances/{}/published_models/{}/learning_iterations'
EVALUATION_METRICS_HREF_PATTERN = u'{}/v3/wml_instances/{}/published_models/{}/evaluation_metrics'
FEEDBACK_HREF_PATTERN = u'{}/v3/wml_instances/{}/published_models/{}/feedback'

DEPLOYMENTS_HREF_PATTERN = u'{}/v4/deployments'
DEPLOYMENT_HREF_PATTERN = u'{}/v4/deployments/{}'
DEPLOYMENT_JOB_HREF_PATTERN = u'{}/v4/deployment_jobs'
DEPLOYMENT_JOBS_HREF_PATTERN = u'{}/v4/deployment_jobs/{}'
DEPLOYMENT_ENVS_HREF_PATTERN = u'{}/v4/deployments/environments'
DEPLOYMENT_ENV_HREF_PATTERN = u'{}/v4/deployments/environments/{}'

MODEL_LAST_VERSION_HREF_PATTERN = u'{}/v4/models/{}'
DEFINITION_HREF_PATTERN = u'{}/v3/ml_assets/training_definitions/{}'
DEFINITIONS_HREF_PATTERN = u'{}/v3/ml_assets/training_definitions'

FUNCTION_HREF_PATTERN = u'{}/v4/functions/{}'
FUNCTION_LATEST_CONTENT_HREF_PATTERN = u'{}/v4/functions/{}/content'
FUNCTIONS_HREF_PATTERN = u'{}/v4/functions'

RUNTIME_HREF_PATTERN = u'{}/v4/runtimes/{}'
RUNTIMES_HREF_PATTERN = u'{}/v4/runtimes'
CUSTOM_LIB_HREF_PATTERN = u'{}/v4/libraries/{}'
CUSTOM_LIBS_HREF_PATTERN = u'{}/v4/libraries'

IAM_TOKEN_API = u'{}&grant_type=urn%3Aibm%3Aparams%3Aoauth%3Agrant-type%3Aapikey'
IAM_TOKEN_URL = u'{}/oidc/token'
PROD_SVT_URL = ['https://us-south.ml.cloud.ibm.com',
                'https://eu-gb.ml.cloud.ibm.com',
                'https://eu-de.ml.cloud.ibm.com',
                'https://jp-tok.ml.cloud.ibm.com',
                'https://ibm-watson-ml.mybluemix.net',
                'https://ibm-watson-ml.eu-gb.bluemix.net',
                'https://private.us-south.ml.cloud.ibm.com',
                'https://private.eu-gb.ml.cloud.ibm.com',
                'https://private.eu-de.ml.cloud.ibm.com',
                'https://private.jp-tok.ml.cloud.ibm.com',
                'https://yp-qa.ml.cloud.ibm.com',
                'https://private.yp-qa.ml.cloud.ibm.com',
                'https://yp-cr.ml.cloud.ibm.com',
                'https://private.yp-cr.ml.cloud.ibm.com']

PIPELINES_HREF_PATTERN=u'{}/v4/pipelines'
PIPELINE_HREF_PATTERN=u'{}/v4/pipelines/{}'


SPACES_HREF_PATTERN = u'{}/v4/spaces'
SPACE_HREF_PATTERN = u'{}/v4/spaces/{}'
MEMBER_HREF_PATTERN=u'{}/v4/spaces/{}/members/{}'
MEMBERS_HREF_PATTERN=u'{}/v4/spaces/{}/members'

SPACES_PLATFORM_HREF_PATTERN = u'{}/v2/spaces'
SPACE_PLATFORM_HREF_PATTERN = u'{}/v2/spaces/{}'
SPACES_MEMBERS_HREF_PATTERN = u'{}/v2/spaces/{}/members'
SPACES_MEMBER_HREF_PATTERN = u'{}/v2/spaces/{}/members/{}'

V4_INSTANCE_ID_HREF_PATTERN = u'{}/ml/v4/instances/{}'

API_VERSION = u'/v4'
SPACES=u'/spaces'
PIPELINES=u'/pipelines'
EXPERIMENTS=u'/experiments'
LIBRARIES=u'/libraries'
RUNTIMES=u'/runtimes'
SOFTWARE_SPEC=u'/software_specifications'
DEPLOYMENTS = u'/deployments'
ASSET = u'{}/v2/assets/{}'
ASSETS = u'{}/v2/assets'
ASSET_TYPE = u'{}/v2/asset_types'
ASSET_FILES = u'{}/v2/asset_files/'
ATTACHMENT = u'{}/v2/assets/{}/attachments/{}'
ATTACHMENT_COMPLETE = u'{}/v2/assets/{}/attachments/{}/complete'
ATTACHMENTS = u'{}/v2/assets/{}/attachments'
SEARCH_ASSETS = u'{}/v2/asset_types/{}/search'
SEARCH_MODEL_DEFINITIONS = u'{}/v2/asset_types/wml_model_definition/search'
SEARCH_DATA_ASSETS = u'{}/v2/asset_types/data_asset/search'
SEARCH_SHINY = u'{}/v2/asset_types/shiny_asset/search'
SEARCH_SCRIPT = u'{}/v2/asset_types/script/search'
GIT_BASED_PROJECT_ASSET = u'{}/userfs/v2/assets/{}'
GIT_BASED_PROJECT_ASSETS = u'{}/userfs/v2/assets'
GIT_BASED_PROJECT_ASSET_TYPE = u'{}/userfs/v2/asset_types'
GIT_BASED_PROJECT_ASSET_FILES = u'{}/v2/asset_files/'
GIT_BASED_PROJECT_ATTACHMENT = u'{}/userfs/v2/assets/{}/attachments/{}'
GIT_BASED_PROJECT_ATTACHMENT_COMPLETE = u'{}/userfs/v2/assets/{}/attachments/{}/complete'
GIT_BASED_PROJECT_ATTACHMENTS = u'{}/userfs/v2/assets/{}/attachments'
GIT_BASED_PROJECT_SEARCH_ASSETS = u'{}/userfs/v2/asset_types/{}/search'
GIT_BASED_PROJECT_SEARCH_MODEL_DEFINITIONS = u'{}/userfs/v2/asset_types/wml_model_definition/search'
GIT_BASED_PROJECT_SEARCH_DATA_ASSETS = u'{}/userfs/v2/asset_types/data_asset/search'
GIT_BASED_PROJECT_SEARCH_SHINY = u'{}/userfs/v2/asset_types/shiny_asset/search'
GIT_BASED_PROJECT_SEARCH_SCRIPT = u'{}/userfs/v2/asset_types/script/search'
DATA_SOURCE_TYPE = u'{}/v2/datasource_types'
DATA_SOURCE_TYPE_BY_ID = u'{}/v2/datasource_types/{}'
CONNECTION_ASSET = u'{}/v2/connections'
CONNECTION_ASSET_SEARCH = u'{}/v2/connections'
CONNECTION_BY_ID = u'{}/v2/connections/{}'
CONNECTIONS_FILES = u'{}/v2/connections/files'
CONNECTIONS_FILE = u'{}/v2/connections/files/{}'
SOFTWARE_SPECIFICATION = u'{}/v2/software_specifications/{}'
SOFTWARE_SPECIFICATIONS = u'{}/v2/software_specifications'
HARDWARE_SPECIFICATION = u'{}/v2/hardware_specifications/{}'
HARDWARE_SPECIFICATIONS = u'{}/v2/hardware_specifications'
PACKAGE_EXTENSION = u'{}/v2/package_extensions/{}'
PACKAGE_EXTENSIONS = u'{}/v2/package_extensions'
PROJECT = u'{}/v2/projects/{}'
LIST_SOFTWARE_SPECIFICATIONS = '{}/ml/v4/list/software_specifications'

V4GA_CLOUD_MIGRATION = u'{}/ml/v4/repository'
V4GA_CLOUD_MIGRATION_ID = u'{}/ml/v4/repository/{}'

REMOTE_TRAINING_SYSTEM = u'{}/v4/remote_training_systems'
REMOTE_TRAINING_SYSTEM_ID = u'{}/v4/remote_training_systems/{}'

EXPORTS = u'{}/v2/asset_exports'
EXPORT_ID = u'{}/v2/asset_exports/{}'
EXPORT_ID_CONTENT = u'{}/v2/asset_exports/{}/content'

IMPORTS = u'{}/v2/asset_imports'
IMPORT_ID = u'{}/v2/asset_imports/{}'

VOLUMES = u'{}/zen-data/v3/service_instances'
VOLUME_ID = u'{}/zen-data/v3/service_instances/{}'
VOLUME_SERVICE = u'{}/zen-data/v1/volumes/volume_services/{}'
VOLUME_SERVICE_FILE_UPLOAD = u'{}/zen-volumes/{}/v1/volumes/files/'

PROMOTE_ASSET = u'{}/projects/api/rest/catalogs/assets/{}/promote'

DATAPLATFORM_URLS_MAP = {
            'https://wml-fvt.ml.test.cloud.ibm.com': 'https://dataplatform.dev.cloud.ibm.com',
            'https://yp-qa.ml.cloud.ibm.com': 'https://dataplatform.test.cloud.ibm.com',
            'https://private.yp-qa.ml.cloud.ibm.com': 'https://dataplatform.test.cloud.ibm.com',
            'https://yp-cr.ml.cloud.ibm.com': 'https://dataplatform.test.cloud.ibm.com',
            'https://private.yp-cr.ml.cloud.ibm.com': 'https://dataplatform.test.cloud.ibm.com',
            'https://jp-tok.ml.cloud.ibm.com': 'https://jp-tok.dataplatform.cloud.ibm.com',
            'https://eu-gb.ml.cloud.ibm.com': 'https://eu-gb.dataplatform.cloud.ibm.com',
            'https://eu-de.ml.cloud.ibm.com': 'https://eu-de.dataplatform.cloud.ibm.com',
            'https://us-south.ml.cloud.ibm.com': 'https://dataplatform.cloud.ibm.com'
        }

WKC_MODEL_REGISTER = u"{}/v1/aigov/model_inventory/models/{}/model_entry"
WKC_MODEL_LIST_FROM_CATALOG = u"{}/v1/aigov/model_inventory/{}/model_entries"
WKC_MODEL_LIST_ALL = u"{}/v1/aigov/model_inventory/model_entries"
TASK_CREDENTIALS = u"{}/v1/task_credentials/{}"
TASK_CREDENTIALS_ALL = u"{}/v1/task_credentials"

def is_url(s):
    res = re.match('https?:\/\/.+', s)
    return res is not None


def is_uid(s):
    res = re.match('[a-z0-9\-]{36}', s)
    return res is not None


class HrefDefinitions:
    def __init__(self, client, cloud_platform_spaces=False, platform_url=None, cams_url=None,
                 cp4d_platform_spaces=False):
        self._wml_credentials = client.wml_credentials
        self._client = client
        self.cloud_platform_spaces = cloud_platform_spaces
        self.cp4d_platform_spaces = cp4d_platform_spaces
        self.platform_url = platform_url
        self.cams_url = cams_url

        if self.cloud_platform_spaces or self.cp4d_platform_spaces:
            self.prepend = '/ml'
        else:
            self.prepend = ''

    def _is_git_based_project(self):
        return self._client.project_type == "local_git_storage"

    def _get_platform_url_if_exists(self):
        return self.platform_url if self.platform_url else self._wml_credentials['url']

    def _get_cams_url_if_exists(self):
        return self.cams_url if self.cams_url else self._wml_credentials['url']

    def get_training_href(self, model_uid):
        return TRAINING_MODEL_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, model_uid)

    def get_trainings_href(self):
        return TRAINING_MODELS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_repo_models_frameworks_href(self):
        return REPO_MODELS_FRAMEWORKS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_instance_endpoint_href(self):
        return INSTANCE_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_instance_by_id_endpoint_href(self):
        return INSTANCE_BY_ID_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'])

    def get_token_endpoint_href(self):
        return TOKEN_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_cpd_token_endpoint_href(self):
        return CPD_TOKEN_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['url'].replace(':31002', ':31843'))

    def get_cpd_bedrock_token_endpoint_href(self):
        return CPD_BEDROCK_TOKEN_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['bedrock_url'])

    def get_cpd_validation_token_endpoint_href(self):
        return CPD_VALIDATION_TOKEN_ENDPOINT_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_published_model_href(self, model_uid):
        return PUBLISHED_MODEL_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, model_uid)

    def get_published_models_href(self):
        return PUBLISHED_MODELS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_learning_configuration_href(self, model_uid):
        return LEARNING_CONFIGURATION_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'], model_uid)

    def get_learning_iterations_href(self, model_uid):
        return LEARNING_ITERATIONS_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'], model_uid)

    def get_learning_iteration_href(self, model_uid, iteration_uid):
        return LEARNING_ITERATION_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'], model_uid, iteration_uid)

    def get_evaluation_metrics_href(self, model_uid):
        return EVALUATION_METRICS_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'], model_uid)

    def get_feedback_href(self, model_uid):
        return FEEDBACK_HREF_PATTERN.format(self._wml_credentials['url'], self._wml_credentials['instance_id'], model_uid)

    def get_model_last_version_href(self, artifact_uid):
        return MODEL_LAST_VERSION_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, artifact_uid)

    def get_deployments_href(self):
        return DEPLOYMENTS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_experiments_href(self):
        return EXPERIMENTS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_experiment_href(self, experiment_uid):
        return EXPERIMENT_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, experiment_uid)

    def get_experiment_runs_href(self, experiment_uid):
        return EXPERIMENT_RUNS_HREF_PATTERN.format(self._wml_credentials['url'], experiment_uid)

    def get_experiment_run_href(self, experiment_uid, experiment_run_uid):
        return EXPERIMENT_RUN_HREF_PATTERN.format(self._wml_credentials['url'], experiment_uid, experiment_run_uid)

    def get_deployment_href(self, deployment_uid):
        return DEPLOYMENT_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, deployment_uid)

    def get_definition_href(self, definition_uid):
        return DEFINITION_HREF_PATTERN.format(self._wml_credentials['url'], definition_uid)

    def get_definitions_href(self):
        return DEFINITIONS_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_function_href(self, ai_function_uid):
        return FUNCTION_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, ai_function_uid)

    def get_function_latest_revision_content_href(self, ai_function_uid):
        return FUNCTION_LATEST_CONTENT_HREF_PATTERN.format(self._wml_credentials['url'], ai_function_uid)

    def get_functions_href(self):
        return FUNCTIONS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_runtime_href_v4(self, runtime_uid):
        return u'/v4/runtimes/{}'.format(runtime_uid)

    def get_runtime_href(self, runtime_uid):
        return RUNTIME_HREF_PATTERN.format(self._wml_credentials['url'], runtime_uid)

    def get_runtimes_href(self):
        return RUNTIMES_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_custom_library_href(self, library_uid):
        return CUSTOM_LIB_HREF_PATTERN.format(self._wml_credentials['url'], library_uid)

    def get_custom_libraries_href(self):
        return CUSTOM_LIBS_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_pipeline_href(self, pipeline_uid):
        return PIPELINE_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, pipeline_uid)

    def get_pipelines_href(self):
        return PIPELINES_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_space_href(self, spaces_uid):
        return SPACE_HREF_PATTERN.format(self._wml_credentials['url'], spaces_uid)

    def get_spaces_href(self):
        return SPACES_HREF_PATTERN.format(self._wml_credentials['url'])

    def get_platform_space_href(self, spaces_id):
        return SPACE_PLATFORM_HREF_PATTERN.format(self._get_platform_url_if_exists(), spaces_id)

    def get_platform_spaces_href(self):
        return SPACES_PLATFORM_HREF_PATTERN.format(self._get_platform_url_if_exists())

    def get_platform_spaces_member_href(self, spaces_id, member_id):
        return SPACES_MEMBER_HREF_PATTERN.format(self._get_platform_url_if_exists(), spaces_id, member_id)

    def get_platform_spaces_members_href(self,spaces_id):
        return SPACES_MEMBERS_HREF_PATTERN.format(self._get_platform_url_if_exists(), spaces_id)

    def get_v4_instance_id_href(self):
        return V4_INSTANCE_ID_HREF_PATTERN.format(self._wml_credentials['url'],
                                                  self._wml_credentials['instance_id'])

    def get_async_deployment_job_href(self):
        return DEPLOYMENT_JOB_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend)

    def get_async_deployment_jobs_href(self, job_uid):
        return DEPLOYMENT_JOBS_HREF_PATTERN.format(self._wml_credentials['url'] + self.prepend, job_uid)

    def get_iam_token_api(self):
        return IAM_TOKEN_API.format(self._wml_credentials['apikey'])

    def get_iam_token_url(self):
        if self._wml_credentials['url'] in PROD_SVT_URL:
            return IAM_TOKEN_URL.format('https://iam.cloud.ibm.com')
        else:
            return IAM_TOKEN_URL.format('https://iam.test.cloud.ibm.com')

    def get_member_href(self, spaces_uid,member_id):
        return MEMBER_HREF_PATTERN.format(self._wml_credentials['url'], spaces_uid,member_id)

    def get_members_href(self,spaces_uid):
        return MEMBERS_HREF_PATTERN.format(self._wml_credentials['url'],spaces_uid)

    def get_data_asset_href(self,asset_id):
        return (ASSET if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET)\
            .format(self._get_cams_url_if_exists(), asset_id)

    def get_data_assets_href(self):
        return (ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSETS)\
            .format(self._get_cams_url_if_exists())

    def get_assets_href(self):
        return (ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSETS)\
            .format(self._get_cams_url_if_exists())

    def get_asset_href(self,asset_id):
        return (ASSET if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET)\
            .format(self._get_cams_url_if_exists(), asset_id)

    def get_base_asset_href(self,asset_id):
        return (ASSET if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET).format('', asset_id)

    def get_base_assets_href(self):
        return (ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSETS).format('')

    def get_base_asset_with_type_href(self,asset_type, asset_id):
        return (ASSET if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET).format('', asset_type) + '/' + asset_id

    def get_attachment_href(self,asset_id,attachment_id):
        return (ATTACHMENT if not self._is_git_based_project() else GIT_BASED_PROJECT_ATTACHMENT)\
            .format(self._get_cams_url_if_exists(), asset_id, attachment_id)

    def get_attachments_href(self, asset_id):
        return (ATTACHMENTS if not self._is_git_based_project() else GIT_BASED_PROJECT_ATTACHMENTS)\
            .format(self._get_cams_url_if_exists(), asset_id)

    def get_attachment_complete_href(self,asset_id,attachment_id):
        return (ATTACHMENT_COMPLETE if not self._is_git_based_project() else GIT_BASED_PROJECT_ATTACHMENT_COMPLETE)\
            .format(self._get_cams_url_if_exists(), asset_id, attachment_id)

    def get_search_asset_href(self):
        return (SEARCH_DATA_ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_SEARCH_DATA_ASSETS)\
            .format(self._get_cams_url_if_exists())

    def get_search_shiny_href(self):
        return (SEARCH_SHINY if not self._is_git_based_project() else GIT_BASED_PROJECT_SEARCH_SHINY)\
            .format(self._get_cams_url_if_exists())

    def get_search_script_href(self):
        return (SEARCH_SCRIPT if not self._is_git_based_project() else GIT_BASED_PROJECT_SEARCH_SCRIPT)\
            .format(self._get_cams_url_if_exists())

    def get_model_definition_assets_href(self):
        return (ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSETS)\
            .format(self._get_cams_url_if_exists())

    def get_model_definition_search_asset_href(self):
        return (SEARCH_MODEL_DEFINITIONS if not self._is_git_based_project() else GIT_BASED_PROJECT_SEARCH_MODEL_DEFINITIONS)\
            .format(self._get_cams_url_if_exists())

    def get_wsd_model_href(self):
        return (ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSETS).format(self._wml_credentials['url'])

    def get_wsd_model_attachment_href(self):
        return (ASSET_FILES if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET_FILES)\
            .format(self._wml_credentials['url'])

    def get_asset_search_href(self, asset_type):
        return (SEARCH_ASSETS if not self._is_git_based_project() else GIT_BASED_PROJECT_SEARCH_ASSETS)\
            .format(self._get_cams_url_if_exists(), asset_type)

    def get_wsd_asset_type_href(self):
        return (ASSET_TYPE if not self._is_git_based_project() else GIT_BASED_PROJECT_ASSET_TYPE)\
            .format(self._wml_credentials['url'])

    def get_wsd_base_href(self):
        return self._wml_credentials['url']

    def get_connections_href(self):
        return CONNECTION_ASSET.format(self._get_platform_url_if_exists())

    def get_connection_by_id_href(self, connection_id):
        return CONNECTION_BY_ID.format(self._get_platform_url_if_exists(), connection_id)

    def get_connections_files_href(self):
        return CONNECTIONS_FILES.format(self._get_platform_url_if_exists())

    def get_connections_file_href(self, file_name):
        return CONNECTIONS_FILE.format(self._get_platform_url_if_exists(), file_name)

    def get_connection_data_types_href(self):
        return DATA_SOURCE_TYPE.format(self._get_platform_url_if_exists())

    def get_sw_spec_href(self, sw_spec_id):
        return SOFTWARE_SPECIFICATION.format(self._get_platform_url_if_exists(), sw_spec_id)

    def get_sw_specs_href(self):
        return SOFTWARE_SPECIFICATIONS.format(self._get_platform_url_if_exists())

    def get_hw_spec_href(self, hw_spec_id):
        return HARDWARE_SPECIFICATION.format(self._get_platform_url_if_exists(), hw_spec_id)

    def get_hw_specs_href(self):
        return HARDWARE_SPECIFICATIONS.format(self._get_platform_url_if_exists())

    def get_pkg_extn_href(self, pkg_extn_id):
        return PACKAGE_EXTENSION.format(self._get_platform_url_if_exists(), pkg_extn_id)

    def get_pkg_extns_href(self):
        return PACKAGE_EXTENSIONS.format(self._get_platform_url_if_exists())

    def get_project_href(self, project_id):
        return PROJECT.format(self._get_platform_url_if_exists(), project_id)

    def get_software_specifications_list_href(self):
        return LIST_SOFTWARE_SPECIFICATIONS.format(self._wml_credentials['url'])

    def v4ga_cloud_migration_href(self):
        return V4GA_CLOUD_MIGRATION.format(self._wml_credentials['url'])

    def v4ga_cloud_migration_id_href(self, migration_id):
        return V4GA_CLOUD_MIGRATION_ID.format(self._wml_credentials['url'], migration_id)

    def exports_href(self):
        return EXPORTS.format(self._get_platform_url_if_exists())

    def export_href(self, export_id):
        return EXPORT_ID.format(self._get_platform_url_if_exists(), export_id)

    def export_content_href(self, export_id):
        return EXPORT_ID_CONTENT.format(self._get_platform_url_if_exists(), export_id)

    def imports_href(self):
        return IMPORTS.format(self._get_platform_url_if_exists())

    def import_href(self, export_id):
        return IMPORT_ID.format(self._get_platform_url_if_exists(), export_id)

    def remote_training_systems_href(self):
        return REMOTE_TRAINING_SYSTEM.format(self._wml_credentials['url'] + self.prepend)

    def remote_training_system_href(self, remote_training_systems_id):
        return REMOTE_TRAINING_SYSTEM_ID.format(self._wml_credentials['url'] + self.prepend, remote_training_systems_id)

    def volumes_href(self):
        return VOLUMES.format(self._wml_credentials['url'])

    def volume_href(self,volume_id):
        return VOLUME_ID.format(self._wml_credentials['url'], volume_id)

    def volume_service_href(self,volume_name):
        return VOLUME_SERVICE.format(self._wml_credentials['url'], volume_name)

    def volume_upload_href(self, volume_name):
        return VOLUME_SERVICE_FILE_UPLOAD.format(self._wml_credentials['url'], volume_name)

    def promote_asset_href(self, asset_id):
        if self.cloud_platform_spaces:
            data_platform_url = DATAPLATFORM_URLS_MAP[self._wml_credentials['url']]
            return PROMOTE_ASSET.format(data_platform_url, asset_id)
        else:
            promote_href = PROMOTE_ASSET.format(self._wml_credentials['url'], asset_id)
            try:
                # note: For CPD older than 4.0 we need to roll back to older endpoint.
                if float(self._wml_credentials.get("version")) < 4.0:
                    promote_href = promote_href.replace("/projects", "")
                # --- end note
            finally:
                return promote_href

    def get_wkc_model_register_href(self, model_id) -> str:
        return WKC_MODEL_REGISTER.format(self._get_platform_url_if_exists(), model_id)

    def get_wkc_model_list_from_catalog_href(self, catalog_id) -> str:
        return WKC_MODEL_LIST_FROM_CATALOG.format(self._get_platform_url_if_exists(), catalog_id)

    def get_wkc_model_list_all_href(self) -> str:
        return WKC_MODEL_LIST_ALL.format(self._get_platform_url_if_exists())

    def get_wkc_model_delete_href(self, asset_id) -> str:
        return WKC_MODEL_REGISTER.format(self._get_platform_url_if_exists(), asset_id)

    def get_task_credentials_href(self, task_credentials_uid) -> str:
        return TASK_CREDENTIALS.format(self._get_platform_url_if_exists(), task_credentials_uid)

    def get_task_credentials_all_href(self) -> str:
        return TASK_CREDENTIALS_ALL.format(self._get_platform_url_if_exists())


    # def get_envs_href(self):
    #     return DEPLOYMENT_ENVS_HREF_PATTERN.format(self._wml_credentials['url'])
    #
    # def get_env_href(self, env_id):
    #     return DEPLOYMENT_ENV_HREF_PATTERN.format(self._wml_credentials['url'],env_id)



"""
 An enumeration class for the message type field which describe party status
"""
from enum import Enum


class StatusType(Enum):
    """
    Status types for Party
    """
    IDLE = 1
    TRAINING = 2
    EVALUATING = 3
    STOPPING = 4



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class PatchOperationExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, op=None, path=None, value=None, _from=None):
        """
        PatchOperationExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'op': 'str',
            'path': 'str',
            'value': 'object',
            '_from': 'str'
        }

        self.attribute_map = {
            'op': 'op',
            'path': 'path',
            'value': 'value',
            '_from': 'from'
        }

        self._op = op
        self._path = path
        self._value = value
        self.__from = _from

    @property
    def op(self):
        """
        Gets the op of this PatchOperationExperiments.
        operation

        :return: The op of this PatchOperationExperiments.
        :rtype: str
        """
        return self._op

    @op.setter
    def op(self, op):
        """
        Sets the op of this PatchOperationExperiments.
        operation

        :param op: The op of this PatchOperationExperiments.
        :type: str
        """
        allowed_values = ["add", "remove", "replace", "move", "copy", "test"]
        if op not in allowed_values:
            raise ValueError(
                "Invalid value for `op` ({0}), must be one of {1}"
                .format(op, allowed_values)
            )

        self._op = op

    @property
    def path(self):
        """
        Gets the path of this PatchOperationExperiments.
        path

        :return: The path of this PatchOperationExperiments.
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """
        Sets the path of this PatchOperationExperiments.
        path

        :param path: The path of this PatchOperationExperiments.
        :type: str
        """

        self._path = path

    @property
    def value(self):
        """
        Gets the value of this PatchOperationExperiments.


        :return: The value of this PatchOperationExperiments.
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Sets the value of this PatchOperationExperiments.


        :param value: The value of this PatchOperationExperiments.
        :type: object
        """

        self._value = value

    @property
    def _from(self):
        """
        Gets the _from of this PatchOperationExperiments.


        :return: The _from of this PatchOperationExperiments.
        :rtype: str
        """
        return self.__from

    @_from.setter
    def _from(self, _from):
        """
        Sets the _from of this PatchOperationExperiments.


        :param _from: The _from of this PatchOperationExperiments.
        :type: str
        """

        self.__from = _from

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



"""
Create routes for the server handler
"""
import logging
import json
from ibmfl.message.message_type import MessageType
from ibmfl.message.message import Message

logger = logging.getLogger(__name__)


def default_end_point(message):
    """
    All Requests whose request type is not determined are routed to this method
    Primary function is to log the request received
    :param request: request received by the server
    :type request: `Message`
    :param response: response to send back
    :type response: `Message`
    :param router: router associated with the request
    :type router: `Router`
    :param kwargs: Dictionary of model-specific arguments.
    :type kwargs: `dict`

    """
    # more detailed logging should be done on request message

    logging.info(message.__dict__)

    message.set_data({'status': 'error'})

    return message


def get_aggregator_router(router, agg_proto_handler):
    """
    Route for register party

    :param router: Router object
    :type router: `Router`
    :param agg_proto_handler: ProtoHandler object
    :type agg_proto_handler: `ProtoHandler`
    :return: None
    """
    router.add_routes({
        '{}'.format(MessageType.REGISTER.value):
        agg_proto_handler.register_party,

        '{}'.format(MessageType.TRAIN.value):
        agg_proto_handler.process_model_update_requests,

        'default': default_end_point
    })


def get_party_router(router, party_proto_handler):
    """
    Route for register party

    :param router: Router object
    :type router: `Router`
    :param party_proto_handler: ProtoHandler object
    :type party_proto_handler: `ProtoHandler`
    :return: None
    """
    router.add_routes({
        '{}'.format(MessageType.TRAIN.value): party_proto_handler.handle_async_request,
        '{}'.format(MessageType.SAVE_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.EVAL_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.SYNC_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.STOP.value): party_proto_handler.handle_request,
        'default': default_end_point
    })

